from __future__ import print_function
from scipy import special, spatial
from sklearn.neighbors import BallTree, NearestNeighbors
from sklearn import metrics
from sklearn.preprocessing import OneHotEncoder, MinMaxScaler
from sklearn.utils.extmath import cartesian
import numpy as np
import math
import warnings

from tigramite.independence_tests.independence_tests_base import CondIndTest

class CMIknnMixed(CondIndTest):
    r"""Conditional mutual information test based on nearest-neighbor estimator.

    Conditional mutual information is a general dependency measure coming
    from an information-theoretic framework. It makes almost no assumptions about the
    parametric form of the dependencies by directly estimating the underlying
    joint density. The tests here are based on the entropy estimation using 
    k-nearest neighbors. We implement three different approaches:
    
    (1) Mesner & Shalizi [1]
    (2) Conditional variant [2]
    (3) Our variant [3]
    
    These approaches differ in how the distance metrics are defined when searching 
    for neighbors: 
    
    (1) The distance on the discrete dimensions for unequal values is 1, otherwise 0.
    (2) This approach splits the space into clusters for which discrete values are 
        all equal, then computes distances between those points (which now have 
        only continuous values). 
    (2) This approach uses the approach from [1], but defines the distance for
        points with unequal discrete values as infinite, and ignores all 
        neighbors that have infinite distances. 

    The tests can be combined with a shuffle test to generate the distribution
    under the null hypothesis of independence, described in [4]. The
    knn-estimator is suitable for heterogeneous variables
    (mixed-type, multivariate with discrete and continuous dimensions). For
    mixture-type variables, use only (1) or (3). 
    
    For continuous variables, use the CMI class. For discrete variables, use
    the CMIsymb or Gsquared class.

    Notes
    -----
    These estimators have as a parameter the number of
    nearest-neighbors :math:`k` which determines the size of hyper-cubes
    around each (high-dimensional) sample point.
    
    For variants (2) and (3), k is used locally, meaning that it defines
    how many neighbors from a respective subsample should be considered.

    :math:`k` can be viewed as a density smoothing parameter (although it is
    data-adaptive unlike fixed-bandwidth estimators). For large :math:`k`, the
    underlying dependencies are more smoothed and CMI has a larger bias,
    but lower variance, which is more important for significance testing. Note
    that the estimated CMI values can be slightly negative while CMI is a non-
    negative quantity.
   
    This method requires the scipy package.

    References
    ----------
    .. [1] Mesner, O.C., & Shalizi, C.R. (2019): Conditional Mutual Information
           Estimation for Mixed Discrete and Continuous Variables with 
           Nearest Neighbors. arXiv: Statistics Theory.
           https://arxiv.org/abs/1912.03387
           
    .. [2] Zan, L., Meynaoui, A., Assaad, C.K., Devijver, E., & Gaussier, 
           É. (2022): A Conditional Mutual Information Estimator for 
           Mixed Data and an Associated Conditional Independence Test. 
           Entropy, 24.
           https://www.mdpi.com/1099-4300/24/9/1234/html

    .. [3] Oana-Iuliana Popescu, Andreas Gerhardus, Martin Rabel, Jakob Runge
           (2024), accepted at CLEAR        
           https://arxiv.org/abs/2310.11132

    .. [4] J. Runge (2018): Conditional Independence Testing Based on a
           Nearest-Neighbor Estimator of Conditional Mutual Information.
           In Proceedings of the 21st International Conference on Artificial
           Intelligence and Statistics.
           http://proceedings.mlr.press/v84/runge18a.html

    Parameters
    ----------
    knn : int or float, optional (default: 0.2)
        Number of nearest-neighbors which determines the size of hyper-cubes
        around each (high-dimensional) sample point. If smaller than 1, this is
        computed as a fraction of T, hence knn=knn*T. For knn larger or equal to
        1, this is the absolute number.

    knn_type : string, optional (default: 'global') 
        Sets the type of heuristic for the MSinf estimator (see paper). Can
        be 'local', 'global', or 'cluster_size'. Use 'global' for the most
        computational efficient variant of the estimator.
        
    estimator : string, optional (default: 'MS')
        The type of estimator to be used. Three options are available:
        approach (1) (Mesner and Shalizi (2021) [1]): 'MS', 
        approach (2) (Zao et.al. (2022) [2]): 'cond',
        approach (3) (Mesner and Shalizi (2021) [1]) with 
        infinite distance for points from different categories): 'FPinf'
        
    shuffle_neighbors : int, optional (default: 5)
        Number of nearest-neighbors within Z for the shuffle surrogates which
        determines the size of hyper-cubes around each (high-dimensional) sample
        point.

    transform : {'ranks', 'standardize',  'scale', False}, optional
        (default: 'standardize')
        Whether to transform the array beforehand by transforming to ranks,
        standardizing or scaling to (a,b)
    
    scale : tuple, optional (default: (0,1))
        the scale (a,b) to use if the transform is set 'scale'
        
    perc : float, optional (default: None)
        the value to be used as percentage of the cluster size for the realization
        of a discrete value when using the 'MSinf' method. If set to None, 
        it is the same as the knn value.

    workers : int (optional, default = -1)
        Number of workers to use for parallel processing. If -1 is given
        all processors are used. Default: -1.
        
    rho: list of float, optional (default: [np.inf])
        Hyperparameters used for weighting the discrete variable distances. 
        If not initialized, the distance will be set to np.inf, such that discrete
        variables with different values will never be considered neighbors. 
        Otherwise the rho
        ...

    significance : str, optional (default: 'shuffle_test')
        Type of significance test to use. For CMIknn only 'fixed_thres' and
        'shuffle_test' are available.

    **kwargs :
        Arguments passed on to parent class CondIndTest.
    """
    @property
    def measure(self):
        """
        Concrete property to return the measure of the independence test
        """
        return self._measure

    def __init__(self,
                 knn=0.2,
                 knn_type='global',
                 estimator='MSinf',
                 shuffle_neighbors=5,
                 significance='shuffle_test',
                 transform='ranks',
                 scale_range=(0, 1),
                 max_with_0=False,
                 workers=-1,
                 **kwargs):
        # Set the member variables
        self.knn = knn
        self.knn_type = knn_type
        self.estimator = estimator
        self.shuffle_neighbors = shuffle_neighbors
        self.transform = transform
        self.max_with_0 = max_with_0
        self.scale_range = scale_range
        self._measure = 'cmi_knn_mixed'
        self.two_sided = False
        self.residual_based = False
        self.recycle_residuals = False
        self.workers = workers
        self.eps = 1e-5
            
        # Call the parent constructor
        CondIndTest.__init__(self, significance=significance, **kwargs)
        # Print some information about construction
        if self.verbosity > 0:
            if self.knn < 1:
                print("knn/T = %s" % self.knn)
            else:
                print("knn = %s" % self.knn)
            print("shuffle_neighbors = %d\n" % self.shuffle_neighbors)
            
    def _standardize_array(self, array, dim):
        """Standardizes a given array with dimensions dim.

        Parameters
        ----------
        array : array-like
            data array with X, Y, Z in rows and observations in columns

        dim: int
            number of dimensions of the data.
        
        Returns
        -------
        array : array-like
            The standardized array.
        """
        array = array.astype(np.float64)
        array -= array.mean(axis=1).reshape(dim, 1)
        std = array.std(axis=1)
        for i in range(dim):
            if std[i] != 0.:
                array[i] /= std[i]
        # array /= array.std(axis=1).reshape(dim, 1)
        # FIXME: If the time series is constant, return nan rather than
        # raising Exception
        if np.any(std == 0.):
            warnings.warn("Possibly constant array!")
            # raise ValueError("nans after standardizing, "
            #                  "possibly constant array!")
        return array
    
    def _scale_array(self, array, minmax=(0, 1)):
        """Scales a given array to range minmax dimension-wise.

        Parameters
        ----------
        array : array-like
            data array with X, Y, Z in rows and observations in columns

        minmax: tuple (a, b)
            the min and the max values (a, b) for the scaling
        
        Returns
        -------
        array : array-like
            The scaled array.
        """
        scaler = MinMaxScaler(minmax)
        return scaler.fit_transform(array.T).T
    
    def _rank_array(self, array):
        """Transform a given array to ranks.

        Parameters
        ----------
        array : array-like
            data array with X, Y, Z in rows and observations in columns
        
        Returns
        -------
        array : array-like
            The scaled array.
        """
        return array.argsort(axis=1).argsort(axis=1).astype(np.float64)
            
    def _transform_mixed_data(self, array, data_type=None, add_noise=True):
        """Applies data transformations to the continuous dimensions of the given data.

        Parameters
        ----------
        array : array-like
            data array with X, Y, Z in rows and observations in columns

        add_noise : bool (default False)
            Defines whether to add small normal noise to the continuous data.
            
        data_type : array-like
            data array of same shape as array which describes whether variables
            are continuous or discrete: 0s for continuous variables and 
            1s for discrete variables

        Returns
        -------
        array : array-like
            The array with the continuous data transformed. 
            
        """
        continuous_idxs = np.where(np.all(data_type == 0, axis=1))[0]  
        cont_dim = len(continuous_idxs)

        if add_noise:
            # Add noise to destroy ties
            array[continuous_idxs, :] += (1E-16 * array[continuous_idxs, :].std(axis=1).reshape(cont_dim, 1)
                  * self.random_state.random((array[continuous_idxs, :].shape[0], array[continuous_idxs, :].shape[1])))
        if self.transform == 'standardize':
            array[continuous_idxs, :] = self._standardize_array(array[continuous_idxs, :], cont_dim)
        elif self.transform == 'scale':
            array[continuous_idxs, :] = self._scale_array(array[continuous_idxs, :], minmax=self.scale_range)
        elif self.transform == 'ranks':
            # if self.estimator == 'MS' or self.estimator == 'FPinf':
            array[continuous_idxs, :] = self._rank_array(array[continuous_idxs, :])
            # else:
                # for conditional, compute ranks for each 
                # pass
        elif self.transform == 'none':
            pass
        else:
            warnings.warn('Unknown transform')
            
        return array
    
    def _transform_to_one_hot_mixed(self, array, xyz, data_type,
                                    zero_inf=False):
        """Applies one-hot encoding to the discrete dimensions of the array.

        Parameters
        ----------
        array : array-like
            data array with X, Y, Z in rows and observations in columns

        xyz : list 
            List that indicates which dimensions belong to which variable, e.g.
            for X, Y, Z one-dimensional xyz = [0, 1, 2]
            
        data_type : array-like
            data array of same shape as array which describes whether variables
            are continuous or discrete: 0s for continuous variables and 
            1s for discrete variables
        
        zero_inf : bool, optional (default: False)
            defines whether to set infinite distances between points with different
            values for the discrete dimensions

        Returns
        -------
        array : array-like
            The array with the continuous data transformed. 
            
        """
        
        discrete_idx_list = np.where(np.all(data_type == 1, axis=0), 1, 0)
        mixed_idx_list = np.where(np.any(data_type == 1, axis=0), 1, 0)

        narray = np.copy(array)
        nxyz = np.copy(xyz)
        ndata_type = np.copy(data_type)

        appended_columns = 0
        for i in range(len(discrete_idx_list)):
            if discrete_idx_list[i] == 1:
                encoder = OneHotEncoder(handle_unknown='ignore')
                i += appended_columns
                data = narray[:, i]
                xyz_val = nxyz[i]
                encoder_df = encoder.fit_transform(data.reshape(-1, 1)).toarray()
                if zero_inf:
                    encoder_df = np.where(encoder_df == 1, 9999999, 0)

                xyz_val = [nxyz[i]] * encoder_df.shape[-1]
                narray = np.concatenate([narray[:, :i], encoder_df, narray[:, i+1:]], axis=-1)

                nxyz = np.concatenate([nxyz[:i], xyz_val, nxyz[i+1:]])
                ndata_type = np.concatenate([ndata_type[:, :i],
                                             np.ones(encoder_df.shape), 
                                             ndata_type[:, i+1:]], 
                                            axis=-1)
                appended_columns += encoder_df.shape[-1] - 1

            elif mixed_idx_list[i] == 1 and zero_inf == True:
                i += appended_columns
                data = narray[:, i]
                xyz_val = nxyz[i]

                # find categories 
                categories = np.unique(narray[:, i] * ndata_type[:, i])
                categories = np.delete(categories, categories == 0.)
                cont_vars = np.unique(narray[:, i] * (1 - ndata_type[:, i]))

                encoder = OneHotEncoder(categories=[categories], handle_unknown='ignore')
                xyz_val = nxyz[i]
                encoder_df = encoder.fit_transform(data.reshape(-1, 1)).toarray()
                if zero_inf:
                    encoder_df = np.where(encoder_df == 1, 9999999, 0)

                xyz_val = [nxyz[i]] * (encoder_df.shape[-1] + 1)
                cont_column = np.expand_dims(narray[:, i] * (1 - ndata_type[:, i]), -1)
                narray = np.concatenate([narray[:, :i], cont_column, encoder_df, narray[:, i+1:]], axis=-1)

                nxyz = np.concatenate([nxyz[:i], xyz_val, nxyz[i+1:]])
                ndata_type = np.concatenate([ndata_type[:, :i], 
                                             np.zeros(cont_column.shape), 
                                             np.ones(encoder_df.shape), 
                                             ndata_type[:, i+1:]], 
                                            axis=-1)
                appended_columns += encoder_df.shape[-1]

        ndiscrete_idx_list = np.where(np.any(ndata_type == 1, axis=0), 1, 0)

        return narray, nxyz, ndata_type, ndiscrete_idx_list         

   
    def get_smallest_cluster_size(self, array, data_type=None):
        """Computes the smallest number of samples for each realization 
        of the discrete variables. 
        Used for computation of the "local" knn. 
        
        Parameters
        ----------
        array : array-like
            data array with X, Y, Z in rows and observations in columns
        
        data_type : array-like
            data array of same shape as array which describes whether variables
            are continuous or discrete: 0s for continuous variables and 
            1s for discrete variables
        Returns
        -------
        min_nc : integer
            The smallest number of samples in a cluster.
        """
        discrete_idx_list = np.where(np.any(data_type == 1, axis=0), 1, 0)
        discrete_xyz_idx = np.where(np.asarray(discrete_idx_list) == 1)[0]
        
        num_xyz_classes = [np.unique(array[:, index]) for index in range(len(discrete_idx_list)) if (discrete_idx_list[index] == 1)]

        xyz_cartesian_product = []

        if len(num_xyz_classes) > 1:
            xyz_cartesian_product = cartesian(num_xyz_classes)
        elif len(num_xyz_classes) > 0:
            xyz_cartesian_product = num_xyz_classes[0]

        min_nc = array.shape[0]
        
        if len(xyz_cartesian_product) > 0:
            for i, entry in enumerate(xyz_cartesian_product):
                current_array = array[np.sum(array[:, discrete_xyz_idx] == entry, 
                                     axis=-1) == len(discrete_xyz_idx)]
                if current_array.shape[0] > 0 and current_array.shape[0] < min_nc:
                    min_nc = current_array.shape[0]

        return min_nc
        
    

    # @jit(forceobj=True)
    def _get_nearest_neighbors_zeroinf_onehot(self, array, xyz, knn,
                                              data_type=None):
        """Returns CMI estimate according to [1] with an
        altered distance metric: the 0-inf metric, which attributes 
        infinite distance to points where the values for the discrete dimensions
        do not coincide.
        
        
        Retrieves the distances eps to the k-th nearest neighbors for every
        sample in joint space XYZ and returns the numbers of nearest neighbors
        within eps in subspaces Z, XZ, YZ. Uses the 0-inf metric for 
        discrete variables.
        
        Parameters
        ----------
        array : array-like
            data array with X, Y, Z in rows and observations in columns

        xyz : array of ints
            XYZ identifier array of shape (dim,).

        knn : int or float
            Number of nearest-neighbors which determines the size of hyper-cubes
            around each (high-dimensional) sample point. If smaller than 1, this
            is computed as a fraction of T, hence knn=knn*T. For knn larger or
            equal to 1, this is the absolute number.
        
        data_type : array-like
            data array of same shape as array which describes whether variables
            are continuous or discrete: 0s for continuous variables and 
            1s for discrete variables
        Returns
        -------
        k_xz, k_yz, k_z : tuple of arrays of shape (T,)
            Nearest neighbors in subspaces.
        """
        dim, T = array.shape

        array = array.astype(np.float64)
        xyz = xyz.astype(np.int32)

        array = self._transform_mixed_data(array, data_type)
        
        array = array.T
        data_type = data_type.T

        narray, nxyz, ndata_type, discrete_idx_list = self._transform_to_one_hot_mixed(array, xyz, data_type,
                                                                                    zero_inf=True)
            
        # Subsample indices
        x_indices = np.where(nxyz == 0)[0]
        y_indices = np.where(nxyz == 1)[0]
        z_indices = np.where(nxyz == 2)[0]
        xz_indices = np.concatenate([x_indices, z_indices])
        yz_indices = np.concatenate([y_indices, z_indices])
     
        # Fit trees
        tree_xyz = spatial.cKDTree(narray)
        neighbors = tree_xyz.query(narray, k=knn+1, p=np.inf,
                                   workers=self.workers,
                                   distance_upper_bound=9999999)
        n, k = neighbors[0].shape
        
        epsarray = np.zeros(n)
        for i in range(n):
            if neighbors[0][i, knn] == np.inf:
                # number of non-inf neighbors
                replacement_idx = np.where(neighbors[0][i] != np.inf)[0][-1]
                if self.knn_type == 'global':
                    # look at at least one neighbor
                    r = max(int(replacement_idx * self.perc), 1)
                elif self.knn_type == 'cluster_size' or self.knn_type == 'local':
                    r = replacement_idx
                epsarray[i] = neighbors[0][i, r]
            else:
                epsarray[i] = neighbors[0][i, knn]
        
        neighbors_radius_xyz = tree_xyz.query_ball_point(narray, epsarray, p=np.inf, workers=self.workers,)

        k_tilde = [len(neighbors_radius_xyz[i]) - 1 if len(neighbors_radius_xyz[i]) > 1 else len(neighbors_radius_xyz[i]) for i in range(len(neighbors_radius_xyz))]
        # k_tilde = [len(neighbors_radius_xyz[i]) for i in range(len(neighbors_radius_xyz))]

        # compute nearest neighbors in subspaces
        xz = narray[:, xz_indices]
        tree_xz = spatial.cKDTree(xz)
        k_xz = tree_xz.query_ball_point(xz, r=epsarray, p=np.inf, workers=self.workers, return_length=True)
        
        yz = narray[:, yz_indices]
        tree_yz = spatial.cKDTree(yz)
        k_yz = tree_yz.query_ball_point(yz, r=epsarray, p=np.inf, workers=self.workers, return_length=True)
            
        if len(z_indices) > 0:
            z = narray[:, z_indices]
            tree_z = spatial.cKDTree(z)
            k_z = tree_z.query_ball_point(z, r=epsarray, p=np.inf, workers=self.workers, return_length=True)
        else:
            # Number of neighbors is T when z is empty.
            k_z = np.full(T, T, dtype='float')

        k_xz = np.asarray([i - 1 if i > 1 else i for i in k_xz])
        k_yz = np.asarray([i - 1 if i > 1 else i for i in k_yz])
        k_z = np.asarray([i - 1 if i > 1 else i for i in k_z])

        return k_tilde, k_xz, k_yz, k_z
    
    def get_dependence_measure_MSinf(self, array, xyz, 
                                       data_type=None):
        """Returns CMI estimate according to Frenzel and Pompe with an
        altered distance metric: the 0-inf metric, which attributes 
        infinite distance to points where the values for the discrete dimensions
        do not coincide.
        
        Parameters
        ----------
        array : array-like
            data array with X, Y, Z in rows and observations in columns

        xyz : array of ints
            XYZ identifier array of shape (dim,).
        
        data_type : array-like
            data array of same shape as array which describes whether variables
            are continuous or discrete: 0s for continuous variables and 
            1s for discrete variables

        Returns
        -------
        val : float
            Conditional mutual information estimate.
        """
        dim, T = array.shape
        
        # compute knn according to knn type
        if self.knn < 1:
            if self.knn_type == 'global':
                # compute knn
                knn = max(1, int(self.knn*T))
                self.perc = self.knn
            elif self.knn_type == 'cluster_size':
                knn = max(1, int(self.knn*T))
            elif self.knn_type == 'local':
                min_nc = self.get_smallest_cluster_size(array.T, data_type.T)
                knn = max(1, int(self.knn*min_nc))
        else:
            if self.knn_type == 'global':
                knn = max(1, int(self.knn))
                self.perc = self.knn
            elif self.knn_type == 'cluster_size':
                knn = max(1, int(self.knn))
            else:
                raise ValueError("MSinf with knn_type == 'local' needs knn value as percentage (value < 1), not number of neighbors!")
        
        knn_tilde, k_xz, k_yz, k_z = self._get_nearest_neighbors_zeroinf_onehot(array=array,
                                                                                 xyz=xyz,
                                                                                 knn=knn,
                                                                                 data_type=data_type)
        non_zero = knn_tilde - k_xz - k_yz + k_z
        
        non_zero_count = np.count_nonzero(non_zero) / len(non_zero)
    
        val = (special.digamma(knn_tilde) - special.digamma(k_xz) -
                                           special.digamma(k_yz) +
                                           special.digamma(k_z))

        val = val[np.isfinite(val)].mean() 
        
        if self.max_with_0 and val < 0.:
            val = 0.
            
        return val
    

    # @jit(forceobj=True)
    def _get_nearest_neighbors_MS_one_hot(self, array, xyz, 
                                          knn, data_type=None):
        """Returns nearest neighbors according to [1].

        Retrieves the distances eps to the k-th nearest neighbors for every
        sample in joint space XYZ and returns the numbers of nearest neighbors
        within eps in subspaces Z, XZ, YZ. Uses a custom-defined metric for 
        discrete variables.

        Parameters
        ----------
        array : array-like
            data array with X, Y, Z in rows and observations in columns

        xyz : array of ints
            XYZ identifier array of shape (dim,).

        knn : int or float
            Number of nearest-neighbors which determines the size of hyper-cubes
            around each (high-dimensional) sample point. If smaller than 1, this
            is computed as a fraction of T, hence knn=knn*T. For knn larger or
            equal to 1, this is the absolute number.
            
        data_type : array-like
            data array of same shape as array which describes whether variables
            are continuous or discrete: 0s for continuous variables and 
            1s for discrete variables

        Returns
        -------
        k_tilde, k_xz, k_yz, k_z : tuple of arrays of shape (T,)
            Nearest neighbors in XYZ, XZ, YZ, and Z subspaces.
        """
        
        dim, T = array.shape
        
        array = array.astype(np.float64)
        xyz = xyz.astype(np.int32)

        array = self._transform_mixed_data(array, data_type)
        
        array = array.T
        data_type = data_type.T
        
        narray, nxyz, ndata_type, discrete_idx_list = self._transform_to_one_hot_mixed(array, 
                                                                                       xyz, 
                                                                                       data_type)
            
        # Subsample indices
        x_indices = np.where(nxyz == 0)[0]
        y_indices = np.where(nxyz == 1)[0]
        z_indices = np.where(nxyz == 2)[0]

        xz_indices = np.concatenate([x_indices, z_indices])
        yz_indices = np.concatenate([y_indices, z_indices])
            
        # Fit trees
        tree_xyz = spatial.cKDTree(narray)
        neighbors = tree_xyz.query(narray, k=knn+1, p=np.inf, workers=self.workers)
        
        epsarray = neighbors[0][:, -1].astype(np.float64)
        
        neighbors_radius_xyz = tree_xyz.query_ball_point(narray, epsarray, p=np.inf, 
                                                         workers=self.workers)
        
        # search again for neighbors in the radius to find all of them
        # in the discrete case k_tilde can be larger than the given knn
        k_tilde = np.asarray([len(neighbors_radius_xyz[i]) - 1 if len(neighbors_radius_xyz[i]) > 1 else len(neighbors_radius_xyz[i]) for i in range(len(neighbors_radius_xyz))])
            
        # compute entropies
        xz = narray[:, xz_indices]
        tree_xz = spatial.cKDTree(xz)
        k_xz = tree_xz.query_ball_point(xz, r=epsarray, p=np.inf,
                                        workers=self.workers, return_length=True)

        
        yz = narray[:, yz_indices]
        tree_yz = spatial.cKDTree(yz)
        k_yz = tree_yz.query_ball_point(yz, r=epsarray, p=np.inf, 
                                        workers=self.workers, return_length=True)

        if len(z_indices) > 0:
            z = narray[:, z_indices]
            tree_z = spatial.cKDTree(z)
            k_z = tree_z.query_ball_point(z, r=epsarray, p=np.inf,
                                          workers=self.workers, return_length=True)
            
        else:
            # Number of neighbors is T when z is empty.
            k_z = np.full(T, T, dtype='float')
        
        k_xz = np.asarray([i - 1 if i > 1 else i for i in k_xz])
        k_yz = np.asarray([i - 1 if i > 1 else i for i in k_yz])
        k_z = np.asarray([i - 1 if i > 1 else i for i in k_z])

        return k_tilde, k_xz, k_yz, k_z
    

    def get_dependence_measure_MS(self, array, xyz,
                                  data_type=None):
        
        """Returns CMI estimate as described in Messner and Shalizi (2021).
        
        Parameters
        ----------
        array : array-like
            data array with X, Y, Z in rows and observations in columns

        xyz : array of ints
            XYZ identifier array of shape (dim,).
        
        data_type : array-like
            data array of same shape as array which describes whether variables
            are continuous or discrete: 0s for continuous variables and 
            1s for discrete variables

        Returns
        -------
        val : float
            Conditional mutual information estimate.
        """
        dim, T = array.shape

        # compute knn
        if self.knn < 1:
            knn = max(1, int(self.knn*T))
        else:
            knn = max(1, self.knn)
        
        
        knn_tilde, k_xz, k_yz, k_z = self._get_nearest_neighbors_MS_one_hot(array=array,
                                                                            xyz=xyz,
                                                                            knn=knn,
                                                                            data_type=data_type)
            
        non_zero = knn_tilde - k_xz - k_yz + k_z
        
        non_zero_count = np.count_nonzero(non_zero) / len(non_zero)
        
        val = (special.digamma(knn_tilde) - special.digamma(k_xz) -
                                           special.digamma(k_yz) +
                                           special.digamma(k_z))
                
        val = val[np.isfinite(val)].mean() 
        
        if self.max_with_0 and val < 0.:
            val = 0.

        return val



    def _compute_entropies_for_discrete_entry(self, array, 
                                          discrete_values, 
                                          discrete_idxs, 
                                          continuous_idxs, 
                                          total_num_samples, 
                                          knn):
        # select data for which the discrete values are the given ones
        current_array = array[np.sum(array[:, discrete_idxs] == discrete_values, 
                                     axis=-1) == len(discrete_idxs)]
        
        # if we do not have samples, we cannot estimate CMI
        if np.size(current_array) == 0:
            return 0., 0.

        T, dim = current_array.shape

        # if we have more samples than knns and samples are not purely discrete, we can
        # compute CMI
        if len(continuous_idxs) > 0 and T > knn:
            val_continuous_entropy = self._compute_continuous_entropy(current_array[:, continuous_idxs], knn)
        else:
            val_continuous_entropy = 0.

        prob = float(T) / total_num_samples

        # multiply by probabilities of occurence
        val_continuous_entropy *= prob
        # compute entropy for that occurence
        val_discrete_entropy = prob * np.log(prob)

        return val_continuous_entropy, val_discrete_entropy

    def _compute_continuous_entropy(self, array, knn):
        T, dim = array.shape
        if T == 1:
            return 0.

        if knn < 1:
            knn = int(max(np.rint(knn * T), 1))
            
        tree = spatial.cKDTree(array)
        epsarray = tree.query(array, k=[knn+1], p=np.inf, 
                              workers=self.workers,
                              eps=0.)[0][:, 0].astype(np.float64)
        epsarray = epsarray[epsarray != 0]
        num_non_zero = len(epsarray)
        if num_non_zero == 0: 
            cmi_hat = 0.
        else:
            avg_dist = float(array.shape[-1]) / float(num_non_zero) * np.sum(np.log(2 * epsarray))
            cmi_hat = special.digamma(num_non_zero) - special.digamma(knn) + avg_dist

        return cmi_hat


    def get_dependence_measure_ZMADG(self, array, xyz,
                                           data_type=None):
        """Returns CMI estimate as described in [2].
        
        Parameters
        ----------
        array : array-like
            data array with X, Y, Z in rows and observations in columns

        xyz : array of ints
            XYZ identifier array of shape (dim,).
            
        data_type : array-like
            data array of same shape as array which describes whether variables
            are continuous or discrete: 0s for continuous variables and 
            1s for discrete variables

        Returns
        -------
        val : float
            Conditional mutual information estimate.
        """
        
        dim, T = array.shape
        
        if self.knn > 1: 
            raise ValueError("ZMADG needs knn value as percentage (value < 1), not number of neighbors!")
        else:
            knn = self.knn
        
        array = array.astype(np.float64)
        xyz = xyz.astype(np.int32)
        
        array = self._transform_mixed_data(array, data_type)
        
        array = array.T
        data_type = data_type.T
        
        discrete_idx_list = np.where(np.any(data_type == 1, axis=0), 1, 0)

        if np.sum(discrete_idx_list) == 0:
            raise ValueError("Variables are continuous, cannot use CMIknnMixed ZMADG!")
            
        # Subsample indices
        x_indices = np.where(xyz == 0)[0]
        y_indices = np.where(xyz == 1)[0]
        z_indices = np.where(xyz == 2)[0]
        xz_indices = np.concatenate([x_indices, z_indices])
        yz_indices = np.concatenate([y_indices, z_indices])

        discrete_xz_indices = discrete_idx_list[xz_indices]
        discrete_yz_indices = discrete_idx_list[yz_indices]
        discrete_z_indices = discrete_idx_list[z_indices]

        discrete_xyz_idx = np.where(np.asarray(discrete_idx_list) == 1)[0]
        discrete_xz_idx = np.where(np.asarray(discrete_xz_indices) == 1)[0]
        discrete_yz_idx = np.where(np.asarray(discrete_yz_indices) == 1)[0]    
        discrete_z_idx = np.where(np.asarray(discrete_z_indices) == 1)[0]  

        continuous_xyz_idx = np.where(np.asarray(discrete_idx_list) == 0)[0]
        continuous_xz_idx = np.where(np.asarray(discrete_xz_indices) == 0)[0]
        continuous_yz_idx = np.where(np.asarray(discrete_yz_indices) == 0)[0]    
        continuous_z_idx = np.where(np.asarray(discrete_z_indices) == 0)[0] 

        # get the number of unique values for each category of the discrete variable
        # add empty set for code not to break when accessing [0]
        num_xz_classes = [np.unique(array[:, xz_indices][:, index]) for index in range(len(discrete_xz_indices)) if (discrete_xz_indices[index] == 1)]
        num_yz_classes = [np.unique(array[:, yz_indices][:, index]) for index in range(len(discrete_yz_indices)) if (discrete_yz_indices[index] == 1)]
        num_z_classes = [np.unique(array[:, z_indices][:, index]) for index in range(len(discrete_z_indices)) if (discrete_z_indices[index] == 1)]
        num_xyz_classes = [np.unique(array[:, index]) for index in range(len(discrete_idx_list)) if (discrete_idx_list[index] == 1)]

        xyz_cartesian_product = []
        xz_cartesian_product = []
        yz_cartesian_product = []
        z_cartesian_product = []

        if len(num_xyz_classes) > 1:
                xyz_cartesian_product = cartesian(num_xyz_classes)
        elif len(num_xyz_classes) > 0:
            xyz_cartesian_product = num_xyz_classes[0]


        if len(num_xz_classes) > 1:
            xz_cartesian_product = cartesian(num_xz_classes)
        elif len(num_xz_classes) > 0:
            xz_cartesian_product = num_xz_classes[0]


        if len(num_yz_classes) > 1:
            yz_cartesian_product = cartesian(num_yz_classes)
        elif len(num_yz_classes) > 0:
            yz_cartesian_product = num_yz_classes[0]


        if len(num_z_classes) > 1:
            z_cartesian_product = cartesian(num_z_classes)
        elif len(num_z_classes) > 0:
            z_cartesian_product = num_z_classes[0]
            
            
        ####### start computing entropies 
        
        if len(xyz_cartesian_product) > 0:
            xyz_cmi = 0.
            xyz_entropy = 0.

            for i, entry in enumerate(xyz_cartesian_product):
                xyz_cont_entropy, xyz_disc_entropy = self._compute_entropies_for_discrete_entry(array, entry,
                                                                       discrete_xyz_idx, 
                                                                       continuous_xyz_idx, 
                                                                       T, knn)
                xyz_cmi += xyz_cont_entropy
                xyz_entropy -= xyz_disc_entropy
        else:
            xyz_cmi = self._compute_continuous_entropy(array, knn)
            xyz_entropy = 0.
            
        h_xyz = xyz_cmi + xyz_entropy

        if len(xz_cartesian_product) > 0:
            xz_cmi = 0.
            xz_entropy = 0.

            for i, entry in enumerate(xz_cartesian_product):
                xz_cont_entropy, xz_disc_entropy = self._compute_entropies_for_discrete_entry(array[:, xz_indices], entry, 
                                                                     discrete_xz_idx, 
                                                                     continuous_xz_idx, 
                                                                     T, knn)
                xz_cmi += xz_cont_entropy
                xz_entropy -= xz_disc_entropy
        else:
            xz_cmi = self._compute_continuous_entropy(array[:, xz_indices], knn)
            xz_entropy = 0.
    
        h_xz = xz_cmi + xz_entropy

        # compute entropies in Xy subspace
        if len(yz_cartesian_product) > 0:
            yz_cmi = 0.
            yz_entropy = 0.

            for i, entry in enumerate(yz_cartesian_product):
                yz_cont_entropy, yz_disc_entropy = self._compute_entropies_for_discrete_entry(array[:, yz_indices], entry, 
                                                                     discrete_yz_idx, 
                                                                     continuous_yz_idx, 
                                                                     T, knn)
                yz_cmi += yz_cont_entropy
                yz_entropy -= yz_disc_entropy
        else:
            yz_cmi = self._compute_continuous_entropy(array[:, yz_indices], knn)
            yz_entropy = 0.

        h_yz = yz_cmi + yz_entropy

        # compute entropies in Z subspace
        if len(z_cartesian_product) > 0:
            z_cmi = 0.
            z_entropy = 0.

            for i, entry in enumerate(z_cartesian_product):
                z_cont_entropy, z_disc_entropy = self._compute_entropies_for_discrete_entry(array[:, z_indices], 
                                                                   entry, 
                                                                   discrete_z_idx, 
                                                                   continuous_z_idx, 
                                                                   T, knn)
                z_cmi += z_cont_entropy
                z_entropy -= z_disc_entropy
        else:
            z_cmi = self._compute_continuous_entropy(array[:, z_indices], knn)
            z_entropy = 0.

        h_z = z_cmi + z_entropy

        # put it all together for the CMI estimation
        val = h_xz + h_yz - h_xyz - h_z

        if self.max_with_0:
            if val < 0.:
                val = 0.
                
        entropies = (xz_cmi, yz_cmi, xyz_cmi, z_cmi, xz_entropy, yz_entropy, xyz_entropy, z_entropy)

        return val
    
    def _get_p_value(self, val, array, xyz, T, dim,
                         data_type=None,
                         sig_override=None):
        """
        Returns the p-value from whichever significance function is specified
        for this test. If an override is used, then it will call a different
        function then specified by self.significance

        Parameters
        ----------
        val : float
            Test statistic value.

        array : array-like
            data array with X, Y, Z in rows and observations in columns

        xyz : array of ints
            XYZ identifier array of shape (dim,).

        T : int
            Sample length

        dim : int
            Dimensionality, ie, number of features.
            
       data_type : array-like
            Binary data array of same shape as array which describes whether 
            individual samples in a variable (or all samples) are continuous 
            or discrete: 0s for continuous variables and 1s for discrete variables.

        sig_override : string
            Must be in 'analytic', 'shuffle_test', 'fixed_thres'

        Returns
        -------
        pval : float or numpy.nan
            P-value.
        """
        # Defaults to the self.significance member value
        use_sig = self.significance
        if sig_override is not None:
            use_sig = sig_override
        # Check if we are using the analytic significance
        if use_sig == 'analytic':
            pval = self.get_analytic_significance(value=val, T=T, dim=dim, xyz=xyz)
        # Check if we are using the shuffle significance
        elif use_sig == 'shuffle_test':
            pval = self.get_shuffle_significance(array=array,
                                                 xyz=xyz,
                                                 value=val, 
                                                 data_type=data_type)
        # Check if we are using the fixed_thres significance
        elif use_sig == 'fixed_thres':
            # Determined outside then
            pval = None
        else:
            raise ValueError("%s not known." % self.significance)
        return pval

    def get_dependence_measure(self, array, xyz, 
                               data_type=None):
        """Calls the appropriate function to estimate CMI.
        Parameters
        ----------
        array : array-like
            data array with X, Y, Z in rows and observations in columns

        xyz : array of ints
            XYZ identifier array of shape (dim,)
            
        data_type : array-like
            data array of same shape as array which describes whether variables
            are continuous or discrete: 0s for continuous variables and 
            1s for discrete variables

        Returns
        -------
        val : float
            Conditional mutual information estimate.
        """
        # check that data is really mixed
        if data_type is None:
            raise ValueError("Type mask cannot be none for CMIknnMixed!")
        if np.sum(data_type) > data_type.size:
            raise ValueError("Type mask contains other values than 0 and 1!")

        if self.estimator == 'MS':
            return self.get_dependence_measure_MS(array,
                                                  xyz,
                                                  data_type)
        elif self.estimator == 'ZMADG':
            return self.get_dependence_measure_ZMADG(array,
                                                           xyz,
                                                           data_type)
        elif self.estimator == 'MSinf':
            return self.get_dependence_measure_MSinf(array,
                                                       xyz,
                                                       data_type)
        else:
            raise ValueError('No such estimator available!')
            
    # @jit(forceobj=True)
    def get_restricted_permutation(self, T, shuffle_neighbors, neighbors, order):

        restricted_permutation = np.zeros(T, dtype=np.int32)
        used = np.array([], dtype=np.int32)

        for sample_index in order:
            # neighbors_to_use = np.unique(neighbors[sample_index, :])
            neighbors_to_use = neighbors[sample_index]
            m = 0
            # use = neighbors[sample_index, m]
            use = neighbors_to_use[m]
            while ((use in used) and (m < shuffle_neighbors - 1)):
                m += 1
                use = neighbors_to_use[m]
                # use = neighbors[sample_index, m]

            restricted_permutation[sample_index] = use
            used = np.append(used, use)

        return restricted_permutation


    # @jit(forceobj=True)
    def _generate_random_permutation(self, array, neighbors, x_indices, data_type):

        T, dim = array.shape
        # Generate random order in which to go through indices loop in
        # next step
        order = self.random_state.permutation(T).astype(np.int32)

        n = np.empty(neighbors.shape[0], dtype=object)

        for i in range(neighbors.shape[0]):
            v = np.unique(neighbors[i])
            # Shuffle neighbor indices for each sample index
            self.random_state.shuffle(v)
            n[i] = v

        # Select a series of neighbor indices that contains as few as
        # possible duplicates
        restricted_permutation = self.get_restricted_permutation(
                T=T,
                shuffle_neighbors=self.shuffle_neighbors,
                neighbors=n,
                order=order)


        array_shuffled = np.copy(array)
        data_type_shuffled = np.copy(data_type)

        for i in x_indices:
            array_shuffled[:, i] = array[restricted_permutation, i]
            data_type_shuffled[:, i] = data_type[restricted_permutation, i]

        return array_shuffled, data_type_shuffled
    
    # @jit(forceobj=True)
    def compute_perm_null_dist(self, array, xyz,
                               data_type=None):
         # max_neighbors = max(1, int(max_neighbor_ratio*T))
        array = self._transform_mixed_data(array.T, data_type.T).T

        # compute valid neighbors
        narray, nxyz, ndata_type, discrete_idx_list = self._transform_to_one_hot_mixed(array, 
                                                                                       xyz, 
                                                                                       data_type,
                                                                                       zero_inf=True)
        x_indices = np.where(nxyz == 0)[0]
        z_indices = np.where(nxyz == 2)[0]

        if self.verbosity > 2:
            print("            nearest-neighbor shuffle significance "
                  "test with n = %d and %d surrogates" % (
                  self.shuffle_neighbors, self.sig_samples))
        # Get nearest neighbors around each sample point in Z
        z_array = np.array(narray[:, z_indices])
        
        tree_xyz = spatial.cKDTree(z_array)
        neighbors = tree_xyz.query(z_array,
                                   k=self.shuffle_neighbors + 1,
                                   p=np.inf,
                                   workers=self.workers,
                                   distance_upper_bound=9999999,
                                   eps=0.)

        # remove all neighbors with distance infinite -> from another class 
        # for those that are discrete 
        valid_neighbors = np.ones(neighbors[1].shape)
        # fill valid neighbors with point -> if infinite, the neighbor will 
        # be the point itself
        valid_neighbors = np.multiply(valid_neighbors, np.expand_dims(np.arange(valid_neighbors.shape[0]), axis=-1))

        valid_neighbors[neighbors[0] != np.inf] = neighbors[1][neighbors[0] != np.inf]
        
        null_dist = np.zeros(self.sig_samples)

        for sam in range(self.sig_samples):
            # permute un-encoded array using the valud neighbors list
            array_shuffled, data_type_shuffled = self._generate_random_permutation(array, 
                                                                                   valid_neighbors, 
                                                                                   x_indices=np.where(xyz == 0)[0],
                                                                                   data_type=data_type)
            
            # use array instead of narray to avoid double encoding
            null_dist[sam] = self.get_dependence_measure(array_shuffled.T,
                                                            xyz,
                                                            data_type=data_type_shuffled.T)
        return null_dist
    
    # @jit(forceobj=True)
    def get_shuffle_significance(self, array, xyz, value,
                                 return_null_dist=False,
                                 data_type=None):
        
        """Returns p-value for nearest-neighbor shuffle significance test.

        For non-empty Z, overwrites get_shuffle_significance from the parent
        class  which is a block shuffle test, which does not preserve
        dependencies of X and Y with Z. Here the parameter shuffle_neighbors is
        used to permute only those values :math:`x_i` and :math:`x_j` for which
        :math:`z_j` is among the nearest neighbors of :math:`z_i`. If Z is
        empty, the block-shuffle test is used.

        Parameters
        ----------
        array : array-like
            data array with X, Y, Z in rows and observations in columns

        xyz : array of ints
            XYZ identifier array of shape (dim,).

        value : number
            Value of test statistic for unshuffled estimate.
        
        data_type : array-like
            data array of same shape as array which describes whether variables
            are continuous or discrete: 0s for continuous variables and 
            1s for discrete variables
        
        Returns
        -------
        pval : float
            p-value
        """
            
        dim, T = array.shape
        array = array.T
        data_type = data_type.T

        z_indices = np.where(xyz == 2)[0]

        if len(z_indices) > 0 and self.shuffle_neighbors < T:
            null_dist = self.compute_perm_null_dist(array, xyz, data_type)
        else:
            null_dist = \
                    self._get_shuffle_dist(array.T, xyz,
                                           sig_samples=self.sig_samples,
                                           sig_blocklength=self.sig_blocklength,
                                           data_type=data_type.T,
                                           verbosity=self.verbosity)

        pval = float(np.sum(null_dist >= value) + 1) / (self.sig_samples + 1)

        if return_null_dist:
            # Sort
            null_dist.sort()
            return pval, null_dist
        return pval

    
    def _get_shuffle_dist(self, array, xyz,
                          sig_samples, sig_blocklength=None,
                          data_type=None,
                          verbosity=0):
        """Returns shuffle distribution of test statistic.

        The rows in array corresponding to the X-variable are shuffled using
        a block-shuffle approach.

        Parameters
        ----------
        array : array-like
            data array with X, Y, Z in rows and observations in columns

        xyz : array of ints
            XYZ identifier array of shape (dim,).

        dependence_measure : object
           Dependence measure function must be of form
           dependence_measure(array, xyz) and return a numeric value

        sig_samples : int, optional (default: 100)
            Number of samples for shuffle significance test.

        sig_blocklength : int, optional (default: None)
            Block length for block-shuffle significance test. If None, the
            block length is determined from the decay of the autocovariance as
            explained in [1]_.
            
        data_type : array-like
            data array of same shape as array which describes whether variables
            are continuous or discrete: 0s for continuous variables and 
            1s for discrete variables

        verbosity : int, optional (default: 0)
            Level of verbosity.

        Returns
        -------
        null_dist : array of shape (sig_samples,)
            Contains the sorted test statistic values estimated from the
            shuffled arrays.
        """
        dim, T = array.shape

        x_indices = np.where(xyz == 0)[0]
        dim_x = len(x_indices)
        
        if sig_blocklength is None:
            sig_blocklength = self._get_block_length(array, xyz,
                                                     mode='significance')

        n_blks = int(math.floor(float(T)/sig_blocklength))
        # print 'n_blks ', n_blks
        if verbosity > 2:
            print("            Significance test with block-length = %d "
                  "..." % (sig_blocklength))

        array_shuffled = np.copy(array)
        data_type_shuffled = np.copy(data_type)
        block_starts = np.arange(0, T - sig_blocklength + 1, sig_blocklength)

        # Dividing the array up into n_blks of length sig_blocklength may
        # leave a tail. This tail is later randomly inserted
        tail = array[x_indices, n_blks*sig_blocklength:]
        tail_type = data_type_shuffled[x_indices, n_blks*sig_blocklength:]
        

        null_dist = np.zeros(sig_samples)
        for sam in range(sig_samples):

            blk_starts = self.random_state.permutation(block_starts)[:n_blks]

            x_shuffled = np.zeros((dim_x, n_blks*sig_blocklength),
                                  dtype=array.dtype)
            type_x_shuffled = np.zeros((dim_x, n_blks*sig_blocklength),
                                  dtype=array.dtype)

            for i, index in enumerate(x_indices):
                for blk in range(sig_blocklength):
                    x_shuffled[i, blk::sig_blocklength] = \
                            array[index, blk_starts + blk]

                    type_x_shuffled[i, blk::sig_blocklength] = \
                            data_type[index, blk_starts + blk]

            # Insert tail randomly somewhere
            if tail.shape[1] > 0:
                insert_tail_at = self.random_state.choice(block_starts)
                x_shuffled = np.insert(x_shuffled, insert_tail_at,
                                       tail.T, axis=1)
                type_x_shuffled = np.insert(type_x_shuffled, insert_tail_at,
                                       tail_type.T, axis=1)

            for i, index in enumerate(x_indices):
                array_shuffled[index] = x_shuffled[i]
                data_type_shuffled[index] = type_x_shuffled[i]
                
            null_dist[sam] = self.get_dependence_measure(array=array_shuffled,
                                                            xyz=xyz,
                                                            data_type=data_type_shuffled)

        return null_dist


if __name__ == '__main__':
    
    import tigramite
    from tigramite.data_processing import DataFrame
    import tigramite.data_processing as pp
    import numpy as np

    from tigramite.independence_tests.cmiknn import CMIknn

    random_state = np.random.default_rng(seed=None)
    cmi = CMIknnMixed(seed=None)

    T = 500
    dimz = 1

    # Continuous data
    z = random_state.standard_normal((T, dimz))
    x = random_state.standard_normal(T).reshape(T, 1)
    y = (5.*z[:,0] + 0.*x[:,0] + random_state.standard_normal(T)).reshape(T, 1)

    print(cmi.get_dependence_measure_raw(x=x,y=y,z=z, 
        x_type=np.zeros(x.shape), y_type=np.zeros(y.shape), z_type=np.zeros(z.shape) ))

    print(cmi.run_test_raw(x=x,y=y,z=z, 
        x_type=np.zeros(x.shape), y_type=np.zeros(y.shape), z_type=np.zeros(z.shape) ))



"""Tigramite causal discovery for time series."""

# Authors: Martin Rabel, Jakob Runge <jakob@jakob-runge.com>
#
# License: GNU General Public License v3.0

import pytest
import numpy as np
from tigramite import data_processing as pp
from tigramite.toymodels import structural_causal_processes as toys
import tigramite.toymodels.non_additive as toy_setup
import tigramite.causal_mediation as mediation
from tigramite.causal_mediation import CausalMediation



"""-------------------------------------------------------------------------------------------
-------------------------------   Test Mixed Data Fit Helpers   ------------------------------
-------------------------------------------------------------------------------------------"""


class ConstantMap:
    def __init__(self, value):
        self.value = value

    def predict(self, x):
        return np.full(np.shape(x)[0], self.value)


class MapAddFirstDim:
    def __init__(self, value):
        self.value = value

    def predict(self, x):
        return np.full(np.shape(x)[0], self.value) + x[:, 0]


def test_mixed_map_0_dim():
    X = toy_setup.CategoricalVariable(categories=2)
    f = mediation.MixedMap([ConstantMap(1), ConstantMap(2)], [X], dtype=np.dtype("float32"))
    r = f({X: np.array([0, 0, 0, 1, 1, 0, 1])})
    assert np.all(r == np.array([1, 1, 1, 2, 2, 1, 2]))


def test_mixed_map_1_dim():
    X = toy_setup.CategoricalVariable(categories=2)
    Z = toy_setup.ContinuousVariable()
    f = mediation.MixedMap([MapAddFirstDim(1), MapAddFirstDim(2)], [X, Z], dtype=np.dtype("float32"))
    r = f({X: np.array([0, 0, 0, 1, 1, 0, 1]), Z: np.arange(7)})
    assert np.all(r == np.array([1, 1, 1, 2, 2, 1, 2] + np.arange(7)))


def test_mixed_fit_pure_cat():
    fit = mediation.FitSetup()
    X = toy_setup.CategoricalVariable(categories=5)
    Y = toy_setup.ContinuousVariable()
    env = toy_setup.Environment(exogenous_noise={
        X: lambda rng, N: rng.binomial(X.categories - 1, 0.7, N),
        Y: lambda rng, N: rng.standard_normal(N)
    }, N=1000, seed=55)
    model = toy_setup.Model(sem={
        X: lambda value: value[X.Noise()],
        Y: lambda value: value[Y.Noise()] + 1.75 * value[X] + value[X] * value[X]
    })

    world_train = toy_setup.World(env, model)
    f = fit.Fit_ContinuousTarget_1D({X: world_train.data[X]}, world_train.data[Y])

    env.ResetWithNewSeed(77)
    world_test = toy_setup.World(env, model)
    err = f({X: world_test.data[X]}) - world_test.data[Y]
    err_m = np.mean(err)
    err_std = np.std(err)
    assert err_m < 0.02, "Mean of prediction-error should be 0, found " + str(err_m)
    assert 0.9 < err_std < 1.1, "Std of prediction-error should be 1 (std of noise on Y), found " + str(err_std)


def test_mixed_fit_true_mixed():
    fit = mediation.FitSetup()
    X1 = toy_setup.CategoricalVariable(categories=5)
    X2 = toy_setup.ContinuousVariable()
    Y = toy_setup.ContinuousVariable()
    env = toy_setup.Environment(exogenous_noise={
        X1: lambda rng, N: rng.binomial(X1.categories - 1, 0.7, N),
        X2: lambda rng, N: rng.standard_normal(N),
        Y: lambda rng, N: rng.standard_normal(N)
    }, N=2000, seed=55)
    model = toy_setup.Model(sem={
        X1: lambda value: value[X1.Noise()],
        X2: lambda value: value[X2.Noise()],
        Y: lambda value: value[Y.Noise()] + 1.75 * value[X1] + value[X1] * value[X1] + np.sin(value[X2])
    })

    world_train = toy_setup.World(env, model)
    f = fit.Fit_ContinuousTarget_1D({X1: world_train.data[X1], X2: world_train.data[X2]}, world_train.data[Y])

    env.ResetWithNewSeed(77)
    world_test = toy_setup.World(env, model)
    err = f({X1: world_test.data[X1], X2: world_test.data[X2]}, ) - world_test.data[Y]
    err_m = np.mean(err)
    err_std = np.std(err)
    assert err_m < 0.05, "Mean of prediction-error should be 0, found " + str(err_m)
    assert 0.9 < err_std < 1.1, "Std of prediction-error should be 1 (std of noise on Y), found " + str(err_std)


def test_mixed_fit_pure_continuous():
    fit = mediation.FitSetup()
    X1 = toy_setup.ContinuousVariable()
    X2 = toy_setup.ContinuousVariable()
    Y = toy_setup.ContinuousVariable()
    env = toy_setup.Environment(exogenous_noise={
        X1: lambda rng, N: rng.standard_normal(N),
        X2: lambda rng, N: rng.standard_normal(N),
        Y: lambda rng, N: rng.standard_normal(N)
    }, N=2000, seed=55)
    model = toy_setup.Model(sem={
        X1: lambda value: value[X1.Noise()],
        X2: lambda value: value[X2.Noise()],
        Y: lambda value: value[Y.Noise()] + 1.75 * value[X1] + value[X1] * value[X1] + np.sin(value[X2])
    })

    world_train = toy_setup.World(env, model)
    f = fit.Fit_ContinuousTarget_1D({X1: world_train.data[X1], X2: world_train.data[X2]}, world_train.data[Y])

    env.ResetWithNewSeed(77)
    world_test = toy_setup.World(env, model)
    err = f({X1: world_test.data[X1], X2: world_test.data[X2]}, ) - world_test.data[Y]
    err_m = np.mean(err)
    err_std = np.std(err)
    assert err_m < 0.05, "Mean of prediction-error should be 0, found " + str(err_m)
    assert 0.9 < err_std < 1.15, "Std of prediction-error should be 1 (std of noise on Y), found " + str(err_std)


def test_mixed_density_pure_cat():
    fit = mediation.FitSetup()
    X = toy_setup.CategoricalVariable(categories=5)
    Y = toy_setup.CategoricalVariable(categories=3)
    env = toy_setup.Environment(exogenous_noise={
        X: lambda rng, N: rng.binomial(X.categories - 1, 0.4, N),
        Y: lambda rng, N: rng.binomial(Y.categories - 1, 0.7, N),
    }, N=1000, seed=55)
    model = toy_setup.Model(sem={
        X: lambda value: value[X.Noise()],
        Y: lambda value: np.clip(value[Y.Noise()] + 1.75 * value[X] + value[X] * value[X],
                                 0, Y.categories-1).astype(dtype="int32")
    })

    world_train = toy_setup.World(env, model)
    f = fit.Fit_CategoricalTarget_MarkovKernel({X: world_train.data[X]}, world_train.data[Y], Y.categories)

    env.ResetWithNewSeed(77)
    world_test = toy_setup.World(env, model)
    predictions = np.mean(f({X: world_test.data[X]}), axis=0)
    print("Predictions: " + str(predictions))
    actual = np.zeros(Y.categories)
    for cY in range(Y.categories):
        actual[cY] = np.count_nonzero(world_test.data[Y] == cY) / env.N
    print("Actual: " + str(actual))
    err_abs_max = np.max(np.abs(predictions - actual))
    assert err_abs_max < 0.01, "Prediction did not agree with ground-truth " + str(np.abs(predictions - actual))


def test_mixed_density_true_mixed():
    fit = mediation.FitSetup()
    X1 = toy_setup.CategoricalVariable(categories=2)
    X2 = toy_setup.ContinuousVariable()
    Y = toy_setup.CategoricalVariable(categories=3)
    env = toy_setup.Environment(exogenous_noise={
        X1: lambda rng, N: rng.binomial(X1.categories - 1, 0.4, N),
        X2: lambda rng, N: rng.standard_normal(N),
        Y: lambda rng, N: rng.binomial(Y.categories - 1, 0.7, N),
    }, N=1000, seed=55)
    model = toy_setup.Model(sem={
        X1: lambda value: value[X1.Noise()],
        X2: lambda value: value[X2.Noise()],
        Y: lambda value: np.clip(value[Y.Noise()] + 1.75 * value[X1] + value[X2] * value[X2],
                                 0, Y.categories-1).astype(dtype="int32")
    })

    world_train = toy_setup.World(env, model)
    f = fit.Fit_CategoricalTarget_MarkovKernel({X1: world_train.data[X1],
                                                X2: world_train.data[X2]}, world_train.data[Y], Y.categories)

    env.ResetWithNewSeed(77)
    world_test = toy_setup.World(env, model)
    predictions = np.mean(f({X1: world_test.data[X1], X2: world_train.data[X2]}), axis=0)
    print("Predictions: " + str(predictions))
    actual = np.zeros(Y.categories)
    for cY in range(Y.categories):
        actual[cY] = np.count_nonzero(world_test.data[Y] == cY) / env.N
    print("Actual: " + str(actual))
    err_abs_max = np.max(np.abs(predictions - actual))
    assert err_abs_max < 0.01, "Prediction did not agree with ground-truth " + str(np.abs(predictions - actual))


def test_mixed_density_true_mixed_xaida_ga_example():
    X = toy_setup.CategoricalVariable("Clouded", categories=2)
    M1 = toy_setup.ContinuousVariable("Temperature")
    M2 = toy_setup.CategoricalVariable("Drought", categories=2)
    Y = toy_setup.CategoricalVariable("Wild-Fire", categories=2)

    env = toy_setup.Environment(exogenous_noise={
        X: lambda rng, N: 1.0 * rng.standard_normal(N),
        M1: lambda rng, N: 1.0 * rng.standard_normal(N),
        M2: lambda rng, N: 1.0 * rng.standard_normal(N),
        Y: lambda rng, N: 1.0 * rng.standard_normal(N)
    }, N=5000, seed=11)
    autocorr = 0.3
    model = toy_setup.Model(sem={
        X: lambda value: autocorr * value[X.Lag(1)] + value[X.Noise()] > 1.0,
        M1: lambda value: autocorr * value[M1.Lag(1)] + (2.0 * value[X.Lag(1)] - 1.0) + value[M1.Noise()],
        M2: lambda value: autocorr * value[M2.Lag(1)] + value[X.Lag(1)] + value[M1.Lag(1)] + value[M2.Noise()] > 1.5,
        Y: lambda value: np.logical_or(autocorr * value[Y.Lag(1)] + value[M1.Lag(1)] + value[Y.Noise()] > 2.5,
                                       np.logical_and(value[M2.Lag(1)],
                                                      autocorr * value[Y.Lag(1)] + 5.0 * value[M1.Lag(1)] + value[
                                                          Y.Noise()] > 2.5))
    })


    fit = mediation.FitSetup(mediation.FitProvider_Continuous_Default.UseSklearn(20))

    world_train = toy_setup.World(env, model)

    M1_past = toy_setup.ContinuousVariable("T_past")
    M2_past = toy_setup.CategoricalVariable("D_past", categories=2)
    Y_past = toy_setup.CategoricalVariable("F_past", categories=2)
    obs = world_train.Observables()
    f = fit.Fit_CategoricalTarget_MarkovKernel({M1_past: obs[M1][0:-1],
                                                 M2_past:   obs[M2][0:-1],
                                                 Y_past:   obs[Y][0:-1]},
                                               obs[Y][1:], Y.categories)


    env.ResetWithNewSeed(77)
    world_test = toy_setup.World(env, model)
    obs = world_test.Observables()
    predictions = np.mean(f({M1_past: obs[M1][0:-1], M2_past: obs[M2][0:-1], Y_past: obs[Y][0:-1]}), axis=0)

    print("Predictions: " + str(predictions))

    actual = np.zeros(Y.categories)
    for cY in range(Y.categories):
        actual[cY] = np.count_nonzero(obs[Y] == cY) / env.N
    print("Actual: " + str(actual))
    err_abs_max = np.max(np.abs(predictions - actual))
    assert err_abs_max < 0.01, "Prediction did not agree with ground-truth " + str(np.abs(predictions - actual))


def test_mixed_density_pure_cont():
    fit = mediation.FitSetup()
    X1 = toy_setup.ContinuousVariable()
    X2 = toy_setup.ContinuousVariable()
    Y = toy_setup.CategoricalVariable(categories=3)
    env = toy_setup.Environment(exogenous_noise={
        X1: lambda rng, N: rng.standard_normal(N),
        X2: lambda rng, N: rng.standard_normal(N),
        Y: lambda rng, N: rng.binomial(Y.categories - 1, 0.7, N),
    }, N=1000, seed=55)
    model = toy_setup.Model(sem={
        X1: lambda value: value[X1.Noise()],
        X2: lambda value: value[X2.Noise()],
        Y: lambda value: np.clip(value[Y.Noise()] + 1.75 * value[X1] + value[X2] * value[X2],
                                 0, Y.categories-1).astype(dtype="int32")
    })

    world_train = toy_setup.World(env, model)
    f = fit.Fit_CategoricalTarget_MarkovKernel({X1: world_train.data[X1],
                                                X2: world_train.data[X2]}, world_train.data[Y], Y.categories)

    env.ResetWithNewSeed(77)
    world_test = toy_setup.World(env, model)
    predictions = np.mean(f({X1: world_test.data[X1], X2: world_train.data[X2]}), axis=0)
    print("Predictions: " + str(predictions))
    actual = np.zeros(Y.categories)
    for cY in range(Y.categories):
        actual[cY] = np.count_nonzero(world_test.data[Y] == cY) / env.N
    print("Actual: " + str(actual))
    err_abs_max = np.max(np.abs(predictions - actual))
    assert err_abs_max < 0.01, "Prediction did not agree with ground-truth " + str(np.abs(predictions - actual))




"""-------------------------------------------------------------------------------------------
-----------------------------   Test Natural Effect Estimators   -----------------------------
-------------------------------------------------------------------------------------------"""



def test_nde():
    A = toy_setup.ContinuousVariable(name="Treatment")
    M = toy_setup.CategoricalVariable(name="Mediator", categories=5)
    Y = toy_setup.ContinuousVariable(name="Target")
    env = toy_setup.Environment(exogenous_noise={
        A: lambda rng, N: rng.standard_normal(N),
        M: lambda rng, N: rng.binomial(M.categories - 1, 0.7, N),
        Y: lambda rng, N: rng.standard_normal(N)
    }, N=10000, seed=23)
    model = toy_setup.Model(sem={
        A: lambda value: value[A.Noise()],
        M: lambda value: np.clip(np.round(value[M.Noise()] + value[A]).astype(dtype=np.dtype("int32")), 0,
                                 M.categories - 1),
        Y: lambda value: value[Y.Noise()] + 1.5 * value[A] + value[M] * value[A] + 0.7 * value[M]
    })

    nde_ground_truth = toy_setup.GroundTruth_NDE(0.0, 1.0, A, Y, M, env, model)
    print("Ground truth for NDE = " + str(nde_ground_truth))

    world = toy_setup.World(env, model)
    nde_estimator = mediation.NaturalEffects_StandardMediationSetup(mediation.FitSetup(), A, Y, M, world.Observables())
    nde_est = nde_estimator.NDE(0.0, 1.0)
    print("Estimation from data for NDE = " + str(nde_est))

    assert np.abs(nde_ground_truth - nde_est) < 0.2, "Estimate of NDE and ground-truth should agree"

    grid = np.arange(-2.0, 2.4, 0.4)
    print("Estimating NDE as function of x on grid: " + str(grid))
    nde_fct_ground_truth = toy_setup.GroundTruth_NDE_fct(A, Y, M, env, model, grid)
    print("Ground truth for NDE(x) = " + str(nde_fct_ground_truth))
    nde_fct_est = nde_estimator.NDE_grid(grid)
    print("Estimation from data for NDE(x) = " + str(nde_fct_est))

    assert np.all(np.abs(nde_fct_ground_truth - nde_fct_est) < 1.3), \
        ("Estimate of NDE as function of x and ground-truth should agree, difference was "
         + str(np.abs(nde_fct_ground_truth - nde_fct_est)))

    nie_ground_truth = toy_setup.GroundTruth_NIE(0.0, 1.0, A, Y, M, env, model)
    print("Ground truth for NIE = " + str(nie_ground_truth))
    nie_est = nde_estimator.NIE(0.0, 1.0)
    print("Estimation from data for NIE = " + str(nie_est))

    assert np.abs(nie_ground_truth - nie_est) < 0.2, "Estimate of NIE and ground-truth should agree"


def test_nde_all_categorical():
    A = toy_setup.CategoricalVariable(name="Treatment", categories=3)
    M = toy_setup.CategoricalVariable(name="Mediator", categories=4)
    Y = toy_setup.CategoricalVariable(name="Target", categories=2)
    env = toy_setup.Environment(exogenous_noise={
        A: lambda rng, N: rng.binomial(A.categories - 1, 0.2, N),
        M: lambda rng, N: rng.binomial(M.categories - 1, 0.5, N),
        Y: lambda rng, N: rng.standard_normal(N)
    }, N=10000, seed=23)
    model = toy_setup.Model(sem={
        A: lambda value: value[A.Noise()],
        M: lambda value: value[M.Noise()],
        # np.clip(np.round(value[M.Noise()] * value[A]), 0, M.categories-1).astype(dtype="int32"),
        Y: lambda value: (value[A] + 2.0 * value[M] + 1.0 * value[Y.Noise()]) > 3.0
    })

    nde_ground_truth = toy_setup.GroundTruth_NDE(1, 2, A, Y, M, env, model)
    print("Ground truth for NDE = " + str(nde_ground_truth))
    nde_change_ground_truth = nde_ground_truth.T[0] - nde_ground_truth.T[1]
    print("Ground truth for NDE change = " + str(nde_change_ground_truth))

    world = toy_setup.World(env, model)
    nde_estimator = mediation.NaturalEffects_StandardMediationSetup(mediation.FitSetup(), A, Y, M, world.Observables())
    nde_est = nde_estimator.NDE(1, 2)
    print("Estimation from data for NDE = " + str(nde_est))

    assert np.all(np.abs(nde_ground_truth - nde_est) < 0.01), \
        "(Indirect) estimate of NDE and ground-truth should agree"

    # Technically there is no reason to treat A as ordinal ...
    grid = np.arange(A.categories - 1)
    nde_ground_truth_fct = toy_setup.GroundTruth_NDE_fct(A, Y, M, env, model, grid, cf_delta=1)
    print("Ground Truth of NDE as function of (ordinal) A is: " + str(nde_ground_truth_fct))

    nde_est_fct = nde_estimator.NDE_grid(grid, cf_delta=1)
    print("Estimate of NDE as function of (ordinal) A is: " + str(nde_est_fct))

    assert np.all(np.abs(nde_ground_truth_fct - nde_est_fct) < 0.01), \
        "Estimate of NDE and ground-truth as function should agree"

    nie_ground_truth_fct = toy_setup.GroundTruth_NIE_fct(A, Y, M, env, model, grid, cf_delta=1)
    print("Ground Truth of NIE as function of (ordinal) A is: " + str(nie_ground_truth_fct))
    nie_est = nde_estimator.NIE_grid(grid, cf_delta=1)
    print("Estimate of NIE as function of (ordinal) A is: " + str(nie_est))

    print("Maximum discrepancy was " + str(np.max(np.abs(nie_ground_truth_fct - nie_est))))

    assert np.all(np.abs(nie_ground_truth_fct - nie_est) < 0.012), \
        "Estimate of NIE and ground-truth as function should agree"


def test_NDE_example2():
    X = toy_setup.ContinuousVariable()
    Y = toy_setup.CategoricalVariable(categories=3)
    M = toy_setup.ContinuousVariable()

    env = toy_setup.Environment(exogenous_noise={
        X: lambda rng, N: 2.0 * rng.standard_normal(N),
        M: lambda rng, N: rng.standard_normal(N),
        Y: lambda rng, N: rng.standard_normal(N)
    }, N=1000, seed=12345)
    model = toy_setup.Model(sem={
        X: lambda value: value[X.Noise()],
        M: lambda value: 0.5 * value[X] + value[M.Noise()],
        Y: lambda value: np.round(np.clip(
            0.2 * value[Y.Noise()] + value[X] + value[M],
            a_min=-0.49, a_max=2.49))
    })

    world = toy_setup.World(env, model)

    fit_setup = mediation.FitSetup()
    estimator = mediation.NaturalEffects_StandardMediationSetup(fit_setup, source=X, target=Y, mediator=M,
                                                          data=world.Observables())

    grid = np.arange(-3.0, 3.6, 0.6)
    nde_fct_est = estimator.NDE_grid(grid, cf_delta=0.6)
    nde_fct_ground_truth = toy_setup.GroundTruth_NDE_fct(X, Y, M, env, model, grid, cf_delta=0.6)

    print("Estimate: " + str(nde_fct_est))
    print("Ground Truth: " + str(nde_fct_ground_truth))
    print("Max. Difference: " + str(np.max(np.abs(nde_fct_est - nde_fct_ground_truth))))
    # Use a liberal limit, as these are a lot (12 * 2 * 3 = 72) values
    assert np.all(np.abs(nde_fct_est - nde_fct_ground_truth) < 0.25), \
        "NDE estimation and ground-truth should agree"

    nie_ground_truth_fct = toy_setup.GroundTruth_NIE_fct(X, Y, M, env, model, grid, cf_delta=0.6)
    print("Ground Truth of NIE as function of (ordinal) A is: " + str(nie_ground_truth_fct))
    nie_est = estimator.NIE_grid(grid, cf_delta=0.6)
    print("Estimate of NIE as function of (ordinal) A is: " + str(nie_est))

    print("Maximum discrepancy was " + str(np.max(np.abs(nie_ground_truth_fct - nie_est))))
    # Use a liberal limit, as these are a lot (12 * 2 * 3 = 72) values
    assert np.all(np.abs(nie_ground_truth_fct - nie_est) < 0.2), \
        "Estimate of NIE and ground-truth as function should agree"


def test_graph_nde():
    X = toy_setup.ContinuousVariable("X")
    M = toy_setup.CategoricalVariable("M", categories=2)
    coeff = 0.8

    env = toy_setup.Environment(exogenous_noise={
        X: lambda rng, N: 1.0 * rng.standard_normal(N),
        M: lambda rng, N: 1.0 * rng.standard_normal(N),
    }, N=5000, seed=11)
    model = toy_setup.Model(sem={
        X: lambda value: value[X.Noise()] + 0.5 * value[M.Lag(1)] + coeff * value[X.Lag(1)],
        M: lambda value: value[X] + value[M.Noise()] + 0.7 * value[M.Lag(1)] > 1.0  # M is binary, ie true or false
    })

    # The Groundtruth Graph
    graph, graph_type = model.GetGroundtruthGraph()

    # Generate data
    world = toy_setup.World(env, model)
    obs = world.Observables()

    # Fit & Estimator Setup
    fit_setup = mediation.FitSetup(mediation.FitProvider_Continuous_Default.UseSklearn(20))
    tau_max = 2

    print("\nAutomatically setting up an estimator")
    estimator = mediation.NaturalEffects_GraphMediation(graph, graph_type, tau_max, fit_setup, world.Observables(),
                                                  effect_source=X.Lag(1), effect_target=X,
                                                  blocked_mediators="all", adjustment_set="auto")
    estimator.PrintInfo()

    print("\nManually setting up an estimator")
    estimator2 = mediation.NaturalEffects_GraphMediation(graph, graph_type, tau_max, fit_setup, world.Observables(),
                                                   effect_source=X.Lag(1), effect_target=X,
                                                   blocked_mediators=[M.Lag(1)], adjustment_set=[M.Lag(2)])
    estimator2.PrintInfo()
    print("")

    # Compute Single-Value Estimates
    nde01 = estimator.NDE(0.0, 1.0)
    nde01_ = estimator2.NDE(0.0, 1.0)
    nde01_gt = toy_setup.GroundTruth_NDE(0.0, 1.0, X.Lag(1), X, [M.Lag(1)], env, model)
    print(f"Estimated NDE at x=0 is {nde01:.3f} (auto) = {nde01_:.3f} (manual), analytical NDE is {coeff:.3f}.")
    print(f"[Numerical Ground-Truth is {nde01_gt:.3f}]")

    # Both estimators should be identical
    assert np.all(nde01 == nde01_)

    print(f"Estimated NDE is {nde01}, analytical NDE is {coeff}, numerical NDE is {nde01_gt}.")
    assert abs(nde01 - coeff) < 0.15  # require est to be somewhat accurate
    assert abs(nde01_gt - coeff) < 0.001  # require numerical est to be much more accurate


def test_graph_nde_fully_binary():
    X = toy_setup.CategoricalVariable("X", categories=2)
    M = toy_setup.CategoricalVariable("M", categories=2)

    env = toy_setup.Environment(exogenous_noise={
        X: lambda rng, N: 1.0 * rng.standard_normal(N),
        M: lambda rng, N: 1.0 * rng.standard_normal(N),
    }, N=5000, seed=11)

    model = toy_setup.Model(sem={
        X: lambda value: value[X.Noise()] + 0.5 * value[M.Lag(1)] + 0.8 * value[X.Lag(1)] > 1.0,
        M: lambda value: value[X] + value[M.Noise()] + 0.7 * value[M.Lag(1)] > 1.0
    })

    graph, graph_type = model.GetGroundtruthGraph()

    world = toy_setup.World(env, model)
    obs = world.Observables()

    fit_setup = mediation.FitSetup()

    tau_max = 2
    estimator = mediation.NaturalEffects_GraphMediation(graph, graph_type, tau_max, fit_setup, obs,
                                                  effect_source=X, effect_target=M,
                                                  blocked_mediators="all", adjustment_set="auto",
                                                  fall_back_to_total_effect=True)
    nde01 = estimator.NDE(False, True)

    nde01_gt = toy_setup.GroundTruth_NDE_auto(False, True, estimator, env, model)

    print(f"Estimated (contemporaneous) NDE is {nde01}, numerical ground-truth is {nde01_gt}.")
    #  Require est to be close to ground-truth
    assert np.all(np.abs(nde01 - nde01_gt) < 0.03)


def test_graph_example_breakout_session():
    X = toy_setup.CategoricalVariable("Clouded", categories=2)
    M1 = toy_setup.ContinuousVariable("Temperature")
    M2 = toy_setup.CategoricalVariable("Drought", categories=2)
    Y = toy_setup.CategoricalVariable("Wild-Fire", categories=2)

    env = toy_setup.Environment(exogenous_noise={
        X: lambda rng, N: 1.0 * rng.standard_normal(N),
        M1: lambda rng, N: 1.0 * rng.standard_normal(N),
        M2: lambda rng, N: 1.0 * rng.standard_normal(N),
        Y: lambda rng, N: 1.0 * rng.standard_normal(N)
    }, N=5000, seed=11)
    model = toy_setup.Model(sem={
        X: lambda value: value[X.Noise()] > 1.0,
        M1: lambda value: (value[X] + 1.0) * value[M1.Noise()],
        M2: lambda value: value[X] + value[M1] + value[M2.Noise()] > 1.0,
        Y: lambda value: np.logical_or(value[M1] + value[Y.Noise()] > 3.0,
                                       np.logical_and(value[M2], value[M1] + value[Y.Noise()] > 1.5))
    })

    graph, graph_type = model.GetGroundtruthGraph()
    graph = graph.reshape(graph.shape[0:2])

    world = toy_setup.World(env, model)

    fit_setup = mediation.FitSetup(mediation.FitProvider_Continuous_Default.UseSklearn(20))
    tau_max = 2
    estimator = mediation.NaturalEffects_GraphMediation(graph, graph_type, tau_max, fit_setup, world.Observables(),
                                                  effect_source=X, effect_target=Y,
                                                  blocked_mediators="all", adjustment_set="auto")

    nde01 = estimator.NDE(False, True)
    nde01_gt = toy_setup.GroundTruth_NDE_auto(False, True, estimator, env, model)

    print(f"Estimated NDE at x=0 is {nde01}, ground-truth is {nde01_gt}.")
    assert np.all(np.abs(nde01 - nde01_gt) < 0.005)




"""-------------------------------------------------------------------------------------------
-------------------------------   Test Non-Additive Toymodels   ------------------------------
-------------------------------------------------------------------------------------------"""




@pytest.fixture(scope="module")
def med_vars():
    class StandardMediationSetup:
        A = toy_setup.ContinuousVariable(name="Treatment")
        M = toy_setup.CategoricalVariable(name="Mediator", categories=5)
        Y = toy_setup.ContinuousVariable(name="Target")

    return StandardMediationSetup()


@pytest.fixture(scope="module")
def env(med_vars):
    return toy_setup.Environment(exogenous_noise={
        med_vars.A: lambda rng, N: rng.standard_normal(N),
        med_vars.M: lambda rng, N: rng.binomial(med_vars.M.categories - 1, 0.7, N),
        med_vars.Y: lambda rng, N: rng.standard_normal(N)
    }, seed=1)


def test_environments(env, med_vars):
    assert len(env.noise) == 3
    assert np.shape(env.GetNoise()[med_vars.A.Noise()]) == (env.N,)
    with pytest.raises(Exception) as e:
        env.Reset()
    assert str(e.value) == ("Cannot reset a fixed-seed environment, use seed=None or factory/lambda instead for"
                            "ensemble-creation.")


def test_model_validation(med_vars):
    A = med_vars.A
    M = med_vars.M
    Y = med_vars.Y
    with pytest.raises(Exception):
        toy_setup.Model(sem={
            A: lambda value: value[A.Noise()],
            Y: lambda value: value[Y.Noise()] + value[A] + value[M],
            M: lambda value: value[M.Noise()] + value[A]
        })


@pytest.fixture(scope="module")
def model(med_vars):
    A = med_vars.A
    M = med_vars.M
    Y = med_vars.Y
    return toy_setup.Model(sem={
        A: lambda value: value[A.Noise()],
        M: lambda value: value[M.Noise()] + value[A],
        Y: lambda value: value[Y.Noise()] + value[A] + value[M]
    })


@pytest.fixture(scope="module")
def model0(model, med_vars):
    return model.Intervene({med_vars.A: 0.0})


@pytest.fixture(scope="module")
def model1(model, med_vars):
    return model.Intervene({med_vars.A: 1.0})


@pytest.fixture(scope="module")
def world(env, model):
    return toy_setup.World(env, model)


@pytest.fixture(scope="module")
def world0(env, model0):
    return toy_setup.World(env, model0)


@pytest.fixture(scope="module")
def world1(env, model1):
    return toy_setup.World(env, model1)


def test_interventions(med_vars, env, world, world0, world1):
    # in the "real" world A is noise, thus not always 0
    assert np.any(world.Observables()[med_vars.A] != 0.0)
    # After intervention A=0, a should always be 0
    assert np.all(world0.Observables()[med_vars.A] == 0.0)
    # After intervention A=0, a should always be 1
    assert np.all(world1.Observables()[med_vars.A] == 1.0)


@pytest.fixture(scope="module")
def cfworld(med_vars, env, model, world, world0, world1):
    cf_world = toy_setup.CounterfactualWorld(env, model)
    cf_world.TakeVariablesFromWorld(world0, [med_vars.M])
    cf_world.TakeVariablesFromWorld(world1, [med_vars.A])
    cf_world.Compute()
    return cf_world


def test_cfworld(med_vars, cfworld, world, world0, world1):
    assert np.all(cfworld.Observables()[med_vars.A] == world1.Observables()[med_vars.A] )
    assert np.all(cfworld.Observables()[med_vars.M] == world0.Observables()[med_vars.M] )
    assert np.all(cfworld.data[med_vars.A.Noise()] == world.data[med_vars.A.Noise()] )
    assert np.all(cfworld.data[med_vars.M.Noise()] == world.data[med_vars.M.Noise()] )
    assert np.all(cfworld.data[med_vars.Y.Noise()] == world.data[med_vars.Y.Noise()] )




"""-------------------------------------------------------------------------------------------
--------------------------------   Test Tigramite Interface   --------------------------------
-------------------------------------------------------------------------------------------"""


def test_tutorial_example0():
    graph =  np.array([[['', '-->', ''],
                    ['', '', ''],
                    ['', '', '']],
                   [['', '-->', ''],
                    ['', '-->', ''],
                    ['-->', '', '-->']],
                   [['', '', ''],
                    ['<--', '', ''],
                    ['', '-->', '']]], dtype='<U3')

    X = [(1,-2)]
    Y = [(2,0)]
    causal_effects = CausalMediation(graph, graph_type='stationary_dag', X=X, Y=Y,
                                    S=None, # (currently S must be None)
                                    hidden_variables=None, # (currently hidden must be None)
                                    verbosity=1)
    var_names = ['$X^0$', '$X^1$', '$X^2$']

    opt = causal_effects.get_optimal_set()

    from tigramite import data_processing as pp
    from tigramite.toymodels import structural_causal_processes as toys

    coeff = .5
    direct_eff = 0.5
    def lin_f(x): return x
    links_coeffs = {
                    0: [((0, -1), coeff, lin_f), ((1, -1), coeff, lin_f)], 
                    1: [((1, -1), coeff, lin_f),], 
                    2: [((2, -1), coeff, lin_f), ((1, 0), coeff, lin_f), ((1,-2), direct_eff, lin_f)],
                    }
    # Observational data
    T = 1000
    data, nonstat = toys.structural_causal_process(links_coeffs, T=T, noises=None, seed=42)
    normalization = []
    data_normalized = np.empty_like(data)
    for v in range(0,3):
        m = np.std(data[:,v])
        normalization.append(m)
        data_normalized[:,v] = data[:,v] / m
    dataframe = pp.DataFrame(data, var_names=var_names)
    dataframe_normalized = pp.DataFrame(data_normalized, var_names=var_names)
    fit_setup = mediation.FitSetup(mediation.FitProvider_Continuous_Default.UseSklearn(20))

    # unnormalized data
    causal_effects.fit_natural_direct_effect(dataframe, blocked_mediators='all',
                                    mixed_data_estimator=fit_setup).PrintInfo()

    nde_est = causal_effects.predict_natural_direct_effect(0.0, 1.0)

    # normalized data
    causal_effects.fit_natural_direct_effect(dataframe_normalized, blocked_mediators='all',
                                    mixed_data_estimator=fit_setup)

    nde_est_from_normalized = causal_effects.predict_natural_direct_effect(0.0, 1.0) * normalization[2] / normalization[1]

    # print results
    print( f"Estimate of the NDE is:\n{nde_est} from unnormalized data, "
        + f"\n{nde_est_from_normalized} from normalized data,\nground-truth is {direct_eff}." )
    
    assert 0.3 < nde_est_from_normalized < 0.6


def test_tutorial_example():    
    graph = np.array([[['', '-->', ''],
                       ['', '', ''],
                       ['', '', '']],
                      [['', '-->', ''],
                       ['', '-->', ''],
                       ['-->', '', '-->']],
                      [['', '', ''],
                       ['<--', '', ''],
                       ['', '-->', '']]], dtype='<U3')

    X = [(1, -2)]
    Y = [(2, 0)]

    causal_effects = CausalMediation(graph, graph_type='stationary_dag', X=X, Y=Y,
                                     S=None,  # (currently S must be None)
                                     hidden_variables=None,  # (currently hidden must be None)
                                     verbosity=1)
    var_names = ['$X^0$', '$X^1$', '$X^2$']


    coeff = .5
    direct_eff = 500.0

    def lin_f(x): return x

    links_coeffs = {
        0: [((0, -1), coeff, lin_f), ((1, -1), coeff, lin_f)],
        1: [((1, -1), coeff, lin_f), ],
        2: [((2, -1), coeff, lin_f), ((1, 0), coeff, lin_f), ((1, -2), direct_eff, lin_f)],
    }
    # Observational data
    T = 10000
    data, nonstat = toys.structural_causal_process(links_coeffs, T=T, noises=None, seed=42)

    normalization = []
    data_normalized = np.empty_like(data)
    for v in range(0, 3):
        m = np.std(data[:, v])
        normalization.append(m)
        data_normalized[:, v] = data[:, v] / m

    dataframe = pp.DataFrame(data_normalized, var_names=var_names)

    fit_setup = mediation.FitSetup(mediation.FitProvider_Continuous_Default.UseSklearn(20))
    causal_effects.fit_natural_direct_effect(dataframe, blocked_mediators='all',
                                             mixed_data_estimator=fit_setup).PrintInfo()
    nde_est = (causal_effects.predict_natural_direct_effect(0.0, 1.0)
               * normalization[2] / normalization[1])
    print(f"Estimate of the NDE is {nde_est}, ground-truth is {direct_eff}.")

    assert 0.8 < nde_est / direct_eff < 1.1




class FitProvider_Continous_Filtered:
    def __init__(self, underlying_fit_provider, filter_to_use):
        self.filter = filter_to_use
        self.underlying_fit_provider = underlying_fit_provider
    def Get_Fit_Continuous(self,x_train,y_train):
        return self.underlying_fit_provider.Get_Fit_Continuous(*self.filter.apply(x_train,y_train))
    
class FitProvider_Density_Filtered:
    def __init__(self, underlying_fit_provider, filter_to_use):
        self.filter = filter_to_use
        self.underlying_fit_provider = underlying_fit_provider
    def Get_Fit_Density(self, x_train):
        return self.underlying_fit_provider.Get_Fit_Density(self.filter.apply(x_train))

class FilterMissingValues:
    def __init__(self, missing_value_flag):
        self.missing_value_flag = missing_value_flag
    def apply(self,x,y=None):
        missing_in_any_x = np.any( x==self.missing_value_flag, axis=1 )
        if y is None:
            valid = np.logical_not( missing_in_any_x )
            return x[valid]
        else:
            missing_in_y = ( y==self.missing_value_flag )
            valid = np.logical_not( np.logical_or(missing_in_any_x, missing_in_y) )
            return x[valid], y[valid]

def apply_filter_to_all_inputs(fit_setup, filter_to_apply):
    # Assume the fit_setup can be contructed from map & density fit and has corresponding members
    # (for all implementations based on the FitSetup class in the mediation-module
    #  of tigramite this is the case; see tutorial on mediation, appendix B)
    return fit_setup.__class__(
        fit_map=FitProvider_Continous_Filtered(fit_setup.fit_map, filter_to_apply),
        fit_density=FitProvider_Density_Filtered(fit_setup.fit_density, filter_to_apply),
    )


def test_tutorial_example_custom_fit():
    graph =  np.array([[['', '-->', ''],
                    ['', '', ''],
                    ['', '', '']],
                   [['', '-->', ''],
                    ['', '-->', ''],
                    ['-->', '', '-->']],
                   [['', '', ''],
                    ['<--', '', ''],
                    ['', '-->', '']]], dtype='<U3')

    X = [(1,-2)]
    Y = [(2,0)]
    var_names = ['$X^0$', '$X^1$', '$X^2$']

    from tigramite import data_processing as pp
    from tigramite.toymodels import structural_causal_processes as toys

    coeff = .5
    direct_eff = 0.5
    def lin_f(x): return x
    links_coeffs = {
                    0: [((0, -1), coeff, lin_f), ((1, -1), coeff, lin_f)], 
                    1: [((1, -1), coeff, lin_f),], 
                    2: [((2, -1), coeff, lin_f), ((1, 0), coeff, lin_f), ((1,-2), direct_eff, lin_f)],
                    }
    # Observational data
    T = 1000
    data, nonstat = toys.structural_causal_process(links_coeffs, T=T, noises=None, seed=None)
    normalization = []
    data_normalized = np.empty_like(data)
    for v in range(0,3):
        m = np.std(data[:,v])
        normalization.append(m)
        data_normalized[:,v] = data[:,v] / m
    dataframe = pp.DataFrame(data, var_names=var_names)
    dataframe_normalized = pp.DataFrame(data_normalized, var_names=var_names)


    seed = 12345
    gap_count = 10
    gap_min_len = 10
    gap_max_len = 20
    rng = np.random.default_rng(seed)
    var_idx = rng.integers(0, data_normalized.shape[1], gap_count)
    offset = rng.integers(0, data_normalized.shape[0]-gap_max_len, gap_count)
    missing_count = rng.integers(10, 20, gap_count)

    modified_data = data_normalized
    for gap in range(gap_count):
        modified_data[offset[gap]:offset[gap]+missing_count[gap], var_idx] = 999

    fit_setup = mediation.FitSetup(mediation.FitProvider_Continuous_Default.UseSklearn(20))
    fit_setup2 = apply_filter_to_all_inputs(fit_setup, FilterMissingValues(999))
    dataframe_unmarked_missing = pp.DataFrame(data_normalized, var_names=var_names) #missing_flag=999)

    causal_effects = CausalMediation(graph, graph_type='stationary_dag', X=X, Y=Y,
                                    S=None, # (currently S must be None)
                                    hidden_variables=None, # (currently hidden must be None)
                                    verbosity=1)
    # normalized data
    causal_effects.fit_natural_direct_effect(dataframe_unmarked_missing, blocked_mediators='all',
                                    mixed_data_estimator=fit_setup2, # set the new fit_setup
                                    enable_dataframe_based_preprocessing=False)

    nde_est = causal_effects.predict_natural_direct_effect(0.0, 1.0) * normalization[2] / normalization[1]

    # print results
    print( f"Estimate of the NDE is:\n{nde_est} with missing values,\nground-truth is {direct_eff}." )


"""Tigramite toymodels."""

# Author: Jakob Runge <jakob@jakob-runge.com>
#
# License: GNU General Public License v3.0
from __future__ import print_function
from collections import defaultdict, OrderedDict
import sys
import warnings
import copy
import math
import numpy as np
import scipy.sparse
import scipy.sparse.linalg
import itertools

def _generate_noise(covar_matrix, time=1000, use_inverse=False):
    """
    Generate a multivariate normal distribution using correlated innovations.

    Parameters
    ----------
    covar_matrix : array
        Covariance matrix of the random variables
    time : int
        Sample size
    use_inverse : bool, optional
        Negate the off-diagonal elements and invert the covariance matrix
        before use

    return_eigenvectors
    -------
    noise : array
        Random noise generated according to covar_matrix
    """
    # Pull out the number of nodes from the shape of the covar_matrix
    n_nodes = covar_matrix.shape[0]
    # Make a deep copy for use in the inverse case
    this_covar = covar_matrix
    # Take the negative inverse if needed
    if use_inverse:
        this_covar = copy.deepcopy(covar_matrix)
        this_covar *= -1
        this_covar[np.diag_indices_from(this_covar)] *= -1
        this_covar = np.linalg.inv(this_covar)
    # Return the noise distribution
    return np.random.multivariate_normal(mean=np.zeros(n_nodes),
                                            cov=this_covar,
                                            size=time)

def _check_stability(graph):
    """
    Raises an AssertionError if the input graph corresponds to a non-stationary
    process.

    Parameters
    ----------
    graph : array
        Lagged connectivity matrices. Shape is (n_nodes, n_nodes, max_delay+1)
    """
    # Get the shape from the input graph
    n_nodes, _, period = graph.shape
    # Set the top section as the horizontally stacked matrix of
    # shape (n_nodes, n_nodes * period)
    stability_matrix = \
        scipy.sparse.hstack([scipy.sparse.lil_matrix(graph[:, :, t_slice])
                             for t_slice in range(period)])
    # Extend an identity matrix of shape
    # (n_nodes * (period - 1), n_nodes * (period - 1)) to shape
    # (n_nodes * (period - 1), n_nodes * period) and stack the top section on
    # top to make the stability matrix of shape
    # (n_nodes * period, n_nodes * period)
    stability_matrix = \
        scipy.sparse.vstack([stability_matrix,
                             scipy.sparse.eye(n_nodes * (period - 1),
                                              n_nodes * period)])
    # Check the number of dimensions to see if we can afford to use a dense
    # matrix
    n_eigs = stability_matrix.shape[0]
    if n_eigs <= 25:
        # If it is relatively low in dimensionality, use a dense array
        stability_matrix = stability_matrix.todense()
        eigen_values, _ = scipy.linalg.eig(stability_matrix)
    else:
        # If it is a large dimensionality, convert to a compressed row sorted
        # matrix, as it may be easier for the linear algebra package
        stability_matrix = stability_matrix.tocsr()
        # Get the eigen values of the stability matrix
        eigen_values = scipy.sparse.linalg.eigs(stability_matrix,
                                                k=(n_eigs - 2),
                                                return_eigenvectors=False)
    # Ensure they all have less than one magnitude
    assert np.all(np.abs(eigen_values) < 1.), \
        "Values given by time lagged connectivity matrix corresponds to a "+\
        " non-stationary process!"

def _check_initial_values(initial_values, shape):
    """
    Raises a AssertionError if the input initial values:
        * Are not a numpy array OR
        * Do not have the shape (n_nodes, max_delay+1)

    Parameters
    ----------
    graph : array
        Lagged connectivity matrices. Shape is (n_nodes, n_nodes, max_delay+1)
    """
    # Ensure it is a numpy array
    assert isinstance(initial_values, np.ndarray),\
        "User must provide initial_values as a numpy.ndarray"
    # Check the shape is correct
    assert initial_values.shape == shape,\
        "Initial values must be of shape (n_nodes, max_delay+1)"+\
        "\n current shape : " + str(initial_values.shape)+\
        "\n desired shape : " + str(shape)

def _var_network(graph,
                 add_noise=True,
                 inno_cov=None,
                 invert_inno=False,
                 T=100,
                 initial_values=None):
    """
    Returns a vector-autoregressive process with correlated innovations.

    Useful for testing.

    Example:
        graph=numpy.array([[[0.2,0.,0.],[0.5,0.,0.]],
                           [[0.,0.1,0. ],[0.3,0.,0.]]])

        represents a process

        X_1(t) = 0.2 X_1(t-1) + 0.5 X_2(t-1) + eps_1(t)
        X_2(t) = 0.3 X_2(t-1) + 0.1 X_1(t-2) + eps_2(t)

        with inv_inno_cov being the negative (except for diagonal) inverse
        covariance matrix of (eps_1(t), eps_2(t)) OR inno_cov being
        the covariance. Initial values can also be provided.


    Parameters
    ----------
    graph : array
        Lagged connectivity matrices. Shape is (n_nodes, n_nodes, max_delay+1)
    add_noise : bool, optional (default: True)
        Flag to add random noise or not
    inno_cov : array, optional (default: None)
        Covariance matrix of innovations.
    invert_inno : bool, optional (defualt : False)
        Flag to negate off-diagonal elements of inno_cov and invert it before
        using it as the covariance matrix of innovations
    T : int, optional (default: 100)
        Sample size.

    initial_values : array, optional (defult: None)
        Initial values for each node. Shape is (n_nodes, max_delay+1), i.e. must
        be of shape (graph.shape[1], graph.shape[2]).

    Returns
    -------
    X : array
        Array of realization.
    """

    n_nodes, _, period = graph.shape

    time = T
    # Test stability
    _check_stability(graph)

    # Generate the returned data
    data = np.random.randn(n_nodes, time)
    # Load the initial values
    if initial_values is not None:
        # Check the shape of the initial values
        _check_initial_values(initial_values, data[:, :period].shape)
        # Input the initial values
        data[:, :period] = initial_values

    # Check if we are adding noise
    noise = None
    if add_noise:
        # Use inno_cov if it was provided
        if inno_cov is not None:
            noise = _generate_noise(inno_cov,
                                    time=time,
                                    use_inverse=invert_inno)
        # Otherwise just use uncorrelated random noise
        else:
            noise = np.random.randn(time, n_nodes)

    for a_time in range(period, time):
        data_past = np.repeat(
            data[:, a_time-period:a_time][:, ::-1].reshape(1, n_nodes, period),
            n_nodes, axis=0)
        data[:, a_time] = (data_past*graph).sum(axis=2).sum(axis=1)
        if add_noise:
            data[:, a_time] += noise[a_time]

    return data.transpose()

def _iter_coeffs(parents_neighbors_coeffs):
    """
    Iterator through the current parents_neighbors_coeffs structure.  Mainly to
    save repeated code and make it easier to change this structure.

    Parameters
    ----------
    parents_neighbors_coeffs : dict
        Dictionary of format:
        {..., j:[((var1, lag1), coef1), ((var2, lag2), coef2), ...], ...} for
        all variables where vars must be in [0..N-1] and lags <= 0 with number
        of variables N.

    Yields
    -------
    (node_id, parent_id, time_lag, coeff) : tuple
        Tuple defining the relationship between nodes across time
    """
    # Iterate through all defined nodes
    for node_id in list(parents_neighbors_coeffs):
        # Iterate over parent nodes and unpack node and coeff
        for (parent_id, time_lag), coeff in parents_neighbors_coeffs[node_id]:
            # Yield the entry
            yield node_id, parent_id, time_lag, coeff

def _check_parent_neighbor(parents_neighbors_coeffs):
    """
    Checks to insure input parent-neighbor connectivity input is sane.  This
    means that:
        * all time lags are non-positive
        * all parent nodes are included as nodes themselves
        * all node indexing is contiguous
        * all node indexing starts from zero
    Raises a ValueError if any one of these conditions are not met.

    Parameters
    ----------
    parents_neighbors_coeffs : dict
        Dictionary of format:
        {..., j:[((var1, lag1), coef1), ((var2, lag2), coef2), ...], ...} for
        all variables where vars must be in [0..N-1] and lags <= 0 with number
        of variables N.
    """
    # Initialize some lists for checking later
    all_nodes = set()
    all_parents = set()
    # Iterate through variables
    for j in list(parents_neighbors_coeffs):
        # Cache all node ids to ensure they are contiguous
        all_nodes.add(j)
    # Iterate through all nodes
    for j, i, tau, _ in _iter_coeffs(parents_neighbors_coeffs):
        # Check all time lags are equal to or less than zero
        if tau > 0:
            raise ValueError("Lag between parent {} and node {}".format(i, j)+\
                             " is {} > 0, must be <= 0!".format(tau))
        # Cache all parent ids to ensure they are mentioned as node ids
        all_parents.add(i)
    # Check that all nodes are contiguous from zero
    all_nodes_list = sorted(list(all_nodes))
    if all_nodes_list != list(range(len(all_nodes_list))):
        raise ValueError("Node IDs in input dictionary must be contiguous"+\
                         " and start from zero!\n"+\
                         " Found IDs : [" +\
                         ",".join(map(str, all_nodes_list))+ "]")
    # Check that all parent nodes are mentioned as a node ID
    if not all_parents.issubset(all_nodes):
        missing_nodes = sorted(list(all_parents - all_nodes))
        all_parents_list = sorted(list(all_parents))
        raise ValueError("Parent IDs in input dictionary must also be in set"+\
                         " of node IDs."+\
                         "\n Parent IDs "+" ".join(map(str, all_parents_list))+\
                         "\n Node IDs "+" ".join(map(str, all_nodes_list)) +\
                         "\n Missing IDs " + " ".join(map(str, missing_nodes)))

def _check_symmetric_relations(a_matrix):
    """
    Check if the argument matrix is symmetric.  Raise a value error with details
    about the offending elements if it is not.  This is useful for checking the
    instantaneously linked nodes have the same link strength.

    Parameters
    ----------
    a_matrix : 2D numpy array
        Relationships between nodes at tau = 0. Indexed such that first index is
        node and second is parent, i.e. node j with parent i has strength
        a_matrix[j,i]
    """
    # Check it is symmetric
    if not np.allclose(a_matrix, a_matrix.T, rtol=1e-10, atol=1e-10):
        # Store the disagreement elements
        bad_elems = ~np.isclose(a_matrix, a_matrix.T, rtol=1e-10, atol=1e-10)
        bad_idxs = np.argwhere(bad_elems)
        error_message = ""
        for node, parent in bad_idxs:
            # Check that we haven't already printed about this pair
            if bad_elems[node, parent]:
                error_message += \
                    "Parent {:d} of node {:d}".format(parent, node)+\
                    " has coefficient {:f}.\n".format(a_matrix[node, parent])+\
                    "Parent {:d} of node {:d}".format(node, parent)+\
                    " has coefficient {:f}.\n".format(a_matrix[parent, node])
            # Check if we already printed about this one
            bad_elems[node, parent] = False
            bad_elems[parent, node] = False
        raise ValueError("Relationships between nodes at tau=0 are not"+\
                         " symmetric!\n"+error_message)

def _find_max_time_lag_and_node_id(parents_neighbors_coeffs):
    """
    Function to find the maximum time lag in the parent-neighbors-coefficients
    object, as well as the largest node ID

    Parameters
    ----------
    parents_neighbors_coeffs : dict
        Dictionary of format:
        {..., j:[((var1, lag1), coef1), ((var2, lag2), coef2), ...], ...} for
        all variables where vars must be in [0..N-1] and lags <= 0 with number
        of variables N.

    Returns
    -------
    (max_time_lag, max_node_id) : tuple
        Tuple of the maximum time lag and maximum node ID
    """
    # Default maximum lag and node ID
    max_time_lag = 0
    max_node_id = len(parents_neighbors_coeffs.keys()) - 1
    # Iterate through the keys in parents_neighbors_coeffs
    for j, _, tau, _ in _iter_coeffs(parents_neighbors_coeffs):
        # Find max lag time
        max_time_lag = max(max_time_lag, abs(tau))
        # Find the max node ID
        # max_node_id = max(max_node_id, j)
    # Return these values
    return max_time_lag, max_node_id

def _get_true_parent_neighbor_dict(parents_neighbors_coeffs):
    """
    Function to return the dictionary of true parent neighbor causal
    connections in time.

    Parameters
    ----------
    parents_neighbors_coeffs : dict
        Dictionary of format:
        {..., j:[((var1, lag1), coef1), ((var2, lag2), coef2), ...], ...} for
        all variables where vars must be in [0..N-1] and lags <= 0 with number
        of variables N.

    Returns
    -------
    true_parent_neighbor : dict
        Dictionary of lists of tuples.  The dictionary is keyed by node ID, the
        list stores the tuple values (parent_node_id, time_lag)
    """
    # Initialize the returned dictionary of lists
    true_parents_neighbors = defaultdict(list)
    for j in parents_neighbors_coeffs:
        for link_props in parents_neighbors_coeffs[j]:
            i, tau = link_props[0]
            coeff = link_props[1]
            # Add parent node id and lag if non-zero coeff
            if coeff != 0.:
                true_parents_neighbors[j].append((i, tau))
    # Return the true relations
    return true_parents_neighbors

def _get_covariance_matrix(parents_neighbors_coeffs):
    """
    Determines the covariance matrix for correlated innovations

    Parameters
    ----------
    parents_neighbors_coeffs : dict
        Dictionary of format:
        {..., j:[((var1, lag1), coef1), ((var2, lag2), coef2), ...], ...} for
        all variables where vars must be in [0..N-1] and lags <= 0 with number
        of variables N.

    Returns
    -------
    covar_matrix : numpy array
        Covariance matrix implied by the parents_neighbors_coeffs.  Used to
        generate correlated innovations.
    """
    # Get the total number of nodes
    _, max_node_id = \
            _find_max_time_lag_and_node_id(parents_neighbors_coeffs)
    n_nodes = max_node_id + 1
    # Initialize the covariance matrix
    covar_matrix = np.identity(n_nodes)
    # Iterate through all the node connections
    for j, i, tau, coeff in _iter_coeffs(parents_neighbors_coeffs):
        # Add to covar_matrix if node connection is instantaneous
        if tau == 0:
            covar_matrix[j, i] = coeff
    return covar_matrix

def _get_lag_connect_matrix(parents_neighbors_coeffs):
    """
    Generates the lagged connectivity matrix from a parent-neighbor
    connectivity dictionary.  Used to generate the input for _var_network

    Parameters
    ----------
    parents_neighbors_coeffs : dict
        Dictionary of format:
        {..., j:[((var1, lag1), coef1), ((var2, lag2), coef2), ...], ...} for
        all variables where vars must be in [0..N-1] and lags <= 0 with number
        of variables N.

    Returns
    -------
    connect_matrix : numpy array
        Lagged connectivity matrix. Shape is (n_nodes, n_nodes, max_delay+1)
    """
    # Get the total number of nodes and time lag
    max_time_lag, max_node_id = \
            _find_max_time_lag_and_node_id(parents_neighbors_coeffs)
    n_nodes = max_node_id + 1
    n_times = max_time_lag + 1
    # Initialize full time graph
    connect_matrix = np.zeros((n_nodes, n_nodes, n_times))
    for j, i, tau, coeff in _iter_coeffs(parents_neighbors_coeffs):
        # If there is a non-zero time lag, add the connection to the matrix
        if tau != 0:
            connect_matrix[j, i, -(tau+1)] = coeff
    # Return the connectivity matrix
    return connect_matrix

def var_process(parents_neighbors_coeffs, T=1000, use='inv_inno_cov',
                verbosity=0, initial_values=None):
    """Returns a vector-autoregressive process with correlated innovations.

    Wrapper around var_network with possibly more user-friendly input options.

    Parameters
    ----------
    parents_neighbors_coeffs : dict
        Dictionary of format: {..., j:[((var1, lag1), coef1), ((var2, lag2),
        coef2), ...], ...} for all variables where vars must be in [0..N-1]
        and lags <= 0 with number of variables N. If lag=0, a nonzero value
        in the covariance matrix (or its inverse) is implied. These should be
        the same for (i, j) and (j, i).
    use : str, optional (default: 'inv_inno_cov')
        Specifier, either 'inno_cov' or 'inv_inno_cov'.
        Any other specifier will result in non-correlated noise.
        For debugging, 'no_noise' can also be specified, in which case random
        noise will be disabled.
    T : int, optional (default: 1000)
        Sample size.
    verbosity : int, optional (default: 0)
        Level of verbosity.
    initial_values : array, optional (default: None)
        Initial values for each node. Shape must be (N, max_delay+1)

    Returns
    -------
    data : array-like
        Data generated from this process
    true_parent_neighbor : dict
        Dictionary of lists of tuples.  The dictionary is keyed by node ID, the
        list stores the tuple values (parent_node_id, time_lag)
    """
    # Check the input parents_neighbors_coeffs dictionary for sanity
    _check_parent_neighbor(parents_neighbors_coeffs)
    # Generate the true parent neighbors graph
    true_parents_neighbors = \
        _get_true_parent_neighbor_dict(parents_neighbors_coeffs)
    # Generate the correlated innovations
    innos = _get_covariance_matrix(parents_neighbors_coeffs)
    # Generate the lagged connectivity matrix for _var_network
    connect_matrix = _get_lag_connect_matrix(parents_neighbors_coeffs)
    # Default values as per 'inno_cov'
    add_noise = True
    invert_inno = False
    # Use the correlated innovations
    if use == 'inno_cov':
        if verbosity > 0:
            print("\nInnovation Cov =\n%s" % str(innos))
    # Use the inverted correlated innovations
    elif use == 'inv_inno_cov':
        invert_inno = True
        if verbosity > 0:
            print("\nInverse Innovation Cov =\n%s" % str(innos))
    # Do not use any noise
    elif use == 'no_noise':
        add_noise = False
        if verbosity > 0:
            print("\nInverse Innovation Cov =\n%s" % str(innos))
    # Use decorrelated noise
    else:
        innos = None
    # Ensure the innovation matrix is symmetric if it is used
    if (innos is not None) and add_noise:
        _check_symmetric_relations(innos)
    # Generate the data using _var_network
    data = _var_network(graph=connect_matrix,
                        add_noise=add_noise,
                        inno_cov=innos,
                        invert_inno=invert_inno,
                        T=T,
                        initial_values=initial_values)
    # Return the data
    return data, true_parents_neighbors

class _Graph():
    r"""Helper class to handle graph properties.

    Parameters
    ----------
    vertices : list
        List of nodes.
    """
    def __init__(self,vertices): 
        self.graph = defaultdict(list) 
        self.V = vertices 
  
    def addEdge(self,u,v):
        """Adding edge to graph."""
        self.graph[u].append(v) 
  
    def isCyclicUtil(self, v, visited, recStack): 
        """Utility function to return whether graph is cyclic."""
        # Mark current node as visited and
        # adds to recursion stack 
        visited[v] = True
        recStack[v] = True
  
        # Recur for all neighbours 
        # if any neighbour is visited and in  
        # recStack then graph is cyclic 
        for neighbour in self.graph[v]: 
            if visited[neighbour] == False: 
                if self.isCyclicUtil(neighbour, visited, recStack) == True: 
                    return True
            elif recStack[neighbour] == True: 
                return True
  
        # The node needs to be poped from  
        # recursion stack before function ends 
        recStack[v] = False
        return False
  
    def isCyclic(self):
        """Returns whether graph is cyclic."""
        visited = [False] * self.V 
        recStack = [False] * self.V 
        for node in range(self.V): 
            if visited[node] == False: 
                if self.isCyclicUtil(node,visited,recStack) == True: 
                    return True
        return False
  
    def topologicalSortUtil(self,v,visited,stack):
        """A recursive function used by topologicalSort ."""
        # Mark the current node as visited.
        visited[v] = True

        # Recur for all the vertices adjacent to this vertex
        for i in self.graph[v]:
            if visited[i] == False:
                self.topologicalSortUtil(i,visited,stack)

        # Push current vertex to stack which stores result
        stack.insert(0,v)

    def topologicalSort(self):
        """A sorting function. """
        # Mark all the vertices as not visited 
        visited = [False]*self.V 
        stack =[] 

        # Call the recursive helper function to store Topological 
        # Sort starting from all vertices one by one 
        for i in range(self.V): 
          if visited[i] == False: 
              self.topologicalSortUtil(i, visited,stack) 

        return stack

def structural_causal_process_ensemble(realizations=10, ensemble_seed=None, **kwargs):
    """Returns an ensemble of time series generated from a structural causal process.

    This adds an ensemble dimension to the output of structural_causal_process.

    See docstring of structural_causal_process for details.

    Parameters
    ----------
    ensemble_seed : int, optional (default: None)
        Random seed for entire ensemble.
    ** kwargs : 
        Arguments of structural_causal_process.

    Returns
    -------
    data : array-like
        Data generated from this process, shape (M, T, N).
    nonvalid : bool
        Indicates whether data has NaNs or infinities.

    """

    nonvalid = False
    for m in range(realizations):

        # Set ensemble seed
        if ensemble_seed is None:
            seed_here = None
        else:
            seed_here = realizations * ensemble_seed + m

        # Get data
        data, nonvalid_here = structural_causal_process(seed=seed_here, **kwargs)
        
        # Update non-validity
        if nonvalid_here:
            nonvalid = True

        if m == 0:
            data_ensemble = np.zeros((realizations,) + data.shape, dtype='float32')

        data_ensemble[m] = data

    return data_ensemble, nonvalid

def structural_causal_process(links, T, noises=None, 
                        intervention=None, intervention_type='hard',
                        transient_fraction=0.2,
                        seed=None):
    """Returns a time series generated from a structural causal process.

    Allows lagged and contemporaneous dependencies and includes the option
    to have intervened variables or particular samples.

    The interventional data is in particular useful for generating ground
    truth for the CausalEffects class.

    In more detail, the method implements a generalized additive noise model process of the form

    .. math:: X^j_t = \\eta^j_t + \\sum_{X^i_{t-\\tau}\\in \\mathcal{P}(X^j_t)}
              c^i_{\\tau} f^i_{\\tau}(X^i_{t-\\tau})

    Links have the format ``{0:[((i, -tau), coeff, func),...], 1:[...],
    ...}`` where ``func`` can be an arbitrary (nonlinear) function provided
    as a python callable with one argument and coeff is the multiplication
    factor. The noise distributions of :math:`\\eta^j` can be specified in
    ``noises``.

    Through the parameters ``intervention`` and ``intervention_type`` the model
    can also be generated with intervened variables.

    Parameters
    ----------
    links : dict
        Dictionary of format: {0:[((i, -tau), coeff, func),...], 1:[...],
        ...} for all variables where i must be in [0..N-1] and tau >= 0 with
        number of variables N. coeff must be a float and func a python
        callable of one argument.
    T : int
        Sample size.
    noises : list of callables or array, optional (default: 'np.random.randn')
        Random distribution function that is called with noises[j](T). If an array,
        it must be of shape ((transient_fraction + 1)*T, N).
    intervention : dict
        Dictionary of format: {1:np.array, ...} containing only keys of intervened
        variables with the value being the array of length T with interventional values.
        Set values to np.nan to leave specific time points of a variable un-intervened.
    intervention_type : str or dict
        Dictionary of format: {1:'hard',  3:'soft', ...} to specify whether intervention is 
        hard (set value) or soft (add value) for variable j. If str, all interventions have 
        the same type.
    transient_fraction : float
        Added percentage of T used as a transient. In total a realization of length
        (transient_fraction + 1)*T will be generated, but then transient_fraction*T will be
        cut off.
    seed : int, optional (default: None)
        Random seed.

    Returns
    -------
    data : array-like
        Data generated from this process, shape (T, N).
    nonvalid : bool
        Indicates whether data has NaNs or infinities.

    """
    random_state = np.random.RandomState(seed)

    N = len(links.keys())
    if noises is None:
        noises = [random_state.randn for j in range(N)]

    if N != max(links.keys())+1:
        raise ValueError("links keys must match N.")

    if isinstance(noises, np.ndarray):
        if noises.shape != (T + int(math.floor(transient_fraction*T)), N):
            raise ValueError("noises.shape must match ((transient_fraction + 1)*T, N).")            
    else:
        if N != len(noises):
            raise ValueError("noises keys must match N.")

    # Check parameters
    max_lag = 0
    contemp_dag = _Graph(N)
    for j in range(N):
        for link_props in links[j]:
            var, lag = link_props[0]
            coeff = link_props[1]
            func = link_props[2]
            if lag == 0: contemp = True
            if var not in range(N):
                raise ValueError("var must be in 0..{}.".format(N-1))
            if 'float' not in str(type(coeff)):
                raise ValueError("coeff must be float.")
            if lag > 0 or type(lag) != int:
                raise ValueError("lag must be non-positive int.")
            max_lag = max(max_lag, abs(lag))

            # Create contemp DAG
            if var != j and lag == 0:
                contemp_dag.addEdge(var, j)

    if contemp_dag.isCyclic() == 1: 
        raise ValueError("Contemporaneous links must not contain cycle.")

    causal_order = contemp_dag.topologicalSort() 

    if intervention is not None:
        if intervention_type is None:
            intervention_type = {j:'hard' for j in intervention}
        elif isinstance(intervention_type, str):
            intervention_type = {j:intervention_type for j in intervention}
        for j in intervention.keys():
            if len(np.atleast_1d(intervention[j])) != T:
                raise ValueError("intervention array for j=%s must be of length T = %d" %(j, T))
            if j not in intervention_type.keys():        
                raise ValueError("intervention_type dictionary must contain entry for %s" %(j))

    transient = int(math.floor(transient_fraction*T))

    data = np.zeros((T+transient, N), dtype='float32')
    for j in range(N):
        if isinstance(noises, np.ndarray):
            data[:, j] = noises[:, j]
        else:
            data[:, j] = noises[j](T+transient)

    for t in range(max_lag, T+transient):
        for j in causal_order:

            if (intervention is not None and j in intervention and t >= transient
                and np.isnan(intervention[j][t - transient]) == False):
                if intervention_type[j] == 'hard':
                    data[t, j] = intervention[j][t - transient]
                    # Move to next j and skip link_props-loop from parents below 
                    continue
                else:
                    data[t, j] += intervention[j][t - transient]

            # This loop is only entered if intervention_type != 'hard'
            for link_props in links[j]:
                var, lag = link_props[0]
                coeff = link_props[1]
                func = link_props[2]
                data[t, j] += coeff * func(data[t + lag, var])

    data = data[transient:]

    nonvalid = (np.any(np.isnan(data)) or np.any(np.isinf(data)))

    return data, nonvalid

def _get_minmax_lag(links):
    """Helper function to retrieve tau_min and tau_max from links.
    """

    N = len(links)

    # Get maximum time lag
    min_lag = np.inf
    max_lag = 0
    for j in range(N):
        for link_props in links[j]:
            if len(link_props) > 2:
                var, lag = link_props[0]
                coeff = link_props[1]
                # func = link_props[2]
                if not isinstance(coeff, float) or coeff != 0.:
                    min_lag = min(min_lag, abs(lag))
                    max_lag = max(max_lag, abs(lag))
            else:
                var, lag = link_props
                min_lag = min(min_lag, abs(lag))
                max_lag = max(max_lag, abs(lag))   

    return min_lag, max_lag

def _get_parents(links, exclude_contemp=False):
    """Helper function to parents from links
    """

    N = len(links)

    # Get maximum time lag
    parents = {}
    for j in range(N):
        parents[j] = []
        for link_props in links[j]:
            var, lag = link_props[0]
            coeff = link_props[1]
            # func = link_props[2]
            if coeff != 0.:
                if not (exclude_contemp and lag == 0):
                    parents[j].append((var, lag))

    return parents

def _get_children(parents):
    """Helper function to children from parents
    """

    N = len(parents)
    children = dict([(j, []) for j in range(N)])

    for j in range(N):
        for par in parents[j]:
            i, tau = par
            children[i].append((j, abs(tau)))

    return children

def links_to_graph(links, tau_max=None):
    """Helper function to convert dictionary of links to graph array format.

    Parameters
    ---------
    links : dict
        Dictionary of form {0:[((0, -1), coeff, func), ...], 1:[...], ...}.
        Also format {0:[(0, -1), ...], 1:[...], ...} is allowed.
    tau_max : int or None
        Maximum lag. If None, the maximum lag in links is used.

    Returns
    -------
    graph : array of shape (N, N, tau_max+1)
        Matrix format of graph with 1 for true links and 0 else.
    """
    N = len(links)

    # Get maximum time lag
    min_lag, max_lag = _get_minmax_lag(links)

    # Set maximum lag
    if tau_max is None:
        tau_max = max_lag
    else:
        if max_lag > tau_max:
            raise ValueError("tau_max is smaller than maximum lag = %d "
                             "found in links, use tau_max=None or larger "
                             "value" % max_lag)

    graph = np.zeros((N, N, tau_max + 1), dtype='<U3')
    for j in links.keys():
        for link_props in links[j]:
            if len(link_props) > 2:
                var, lag = link_props[0]
                coeff = link_props[1]
                if coeff != 0.:
                    graph[var, j, abs(lag)] = "-->"
                    if lag == 0:
                        graph[j, var, 0] = "<--"
            else:
                var, lag = link_props
                graph[var, j, abs(lag)] = "-->"
                if lag == 0:
                    graph[j, var, 0] = "<--"

    return graph

def dag_to_links(dag):
    """Helper function to convert DAG graph to dictionary of parents.

    Parameters
    ---------
    dag : array of shape (N, N, tau_max+1)
        Matrix format of graph in string format. Must be DAG.

    Returns
    -------
    parents : dict
        Dictionary of form {0:[(0, -1), ...], 1:[...], ...}.
    """
    N = dag.shape[0]

    parents = dict([(j, []) for j in range(N)])

    allowed_edges = ["-->", "<--", ""] 
    if np.any(np.isin(dag, allowed_edges) == False):
        raise ValueError("graph must be a DAG.")

    for (i, j, tau) in zip(*np.where(dag=='-->')):
        parents[j].append((i, -tau))

    return parents

def generate_structural_causal_process(
                N=2, 
                L=1, 
                dependency_funcs=['linear'], 
                dependency_coeffs=[-0.5, 0.5], 
                auto_coeffs=[0.5, 0.7], 
                contemp_fraction=0.,
                max_lag=1, 
                noise_dists=['gaussian'],
                noise_means=[0.],
                noise_sigmas=[0.5, 2.],
                noise_seed=None,
                seed=None):
    """"Randomly generates a structural causal process based on input characteristics.

    The process has the form 

    .. math:: X^j_t = \\eta^j_t + a^j X^j_{t-1} + \\sum_{X^i_{t-\\tau}\\in pa(X^j_t)}
              c^i_{\\tau} f^i_{\\tau}(X^i_{t-\\tau})

    where ``j = 1, ..., N``. Here the properties of :math:`\\eta^j_t` are
    randomly frawn from the noise parameters (see below), :math:`pa
    (X^j_t)` are the causal parents drawn randomly such that in total ``L``
    links occur out of which ``contemp_fraction`` are contemporaneous and
    their time lags are drawn from ``[0 or 1..max_lag]``, the
    coefficients :math:`c^i_{\\tau}` are drawn from
    ``dependency_coeffs``, :math:`a^j` are drawn from ``auto_coeffs``,
    and :math:`f^i_{\\tau}` are drawn from ``dependency_funcs``.

    The returned dictionary links has the format 
    ``{0:[((i, -tau), coeff, func),...], 1:[...], ...}`` 
    where ``func`` can be an arbitrary (nonlinear) function provided
    as a python callable with one argument and coeff is the multiplication
    factor. The noise distributions of :math:`\\eta^j` are returned in
    ``noises``, see specifics below.

    The process might be non-stationary. In case of asymptotically linear
    dependency functions and no contemporaneous links this can be checked with
    ``check_stationarity(...)``. Otherwise check by generating a large sample
    and test for np.inf.

    Parameters
    ---------
    N : int
        Number of variables.
    L : int
        Number of cross-links between two different variables.
    dependency_funcs : list
        List of callables or strings 'linear' or 'nonlinear' for a linear and a specific nonlinear function
        that is asymptotically linear.
    dependency_coeffs : list
        List of floats from which the coupling coefficients are randomly drawn.
    auto_coeffs : list
        List of floats from which the lag-1 autodependencies are randomly drawn.
    contemp_fraction : float [0., 1]
        Fraction of the L links that are contemporaneous (lag zero).
    max_lag : int
        Maximum lag from which the time lags of links are drawn.
    noise_dists : list
        List of noise functions. Either in
        {'gaussian', 'weibull', 'uniform'} or user-specified, in which case
        it must be parametrized just by the size parameter. E.g. def beta
        (T): return np.random.beta(a=1, b=0.5, T)
    noise_means : list
        Noise mean. Only used for noise in {'gaussian', 'weibull', 'uniform'}.
    noise_sigmas : list
        Noise standard deviation. Only used for noise in {'gaussian', 'weibull', 'uniform'}.   
    seed : int
        Random seed to draw the above random functions from.
    noise_seed : int
        Random seed for noise function random generator.

    Returns
    -------
    links : dict
        Dictionary of form {0:[((0, -1), coeff, func), ...], 1:[...], ...}.
    noises : list
        List of N noise functions to call by noise(T) where T is the time series length.
    """
    
    # Init random states
    random_state = np.random.RandomState(seed)
    random_state_noise = np.random.RandomState(noise_seed)

    def linear(x): return x
    def nonlinear(x): return (x + 5. * x**2 * np.exp(-x**2 / 20.))

    if max_lag == 0:
        contemp_fraction = 1.

    if contemp_fraction > 0.:
        ordered_pairs = list(itertools.combinations(range(N), 2))
        max_poss_links = min(L, len(ordered_pairs))
        L_contemp = int(contemp_fraction*max_poss_links)
        L_lagged = max_poss_links - L_contemp
    else:
        L_lagged = L
        L_contemp = 0

    # Random order
    causal_order = list(random_state.permutation(N))

    # Init link dict
    links = dict([(i, []) for i in range(N)])

    # Generate auto-dependencies at lag 1
    if max_lag > 0:
        for i in causal_order:
            a = random_state.choice(auto_coeffs)
            if a != 0.:
                links[i].append(((int(i), -1), float(a), linear))

    # Non-cyclic contemp random pairs of links such that
    # index of cause < index of effect
    # Take up to (!) L_contemp links
    ordered_pairs = list(itertools.combinations(range(N), 2))
    random_state.shuffle(ordered_pairs)
    contemp_links = [(causal_order[pair[0]], causal_order[pair[1]]) 
                    for pair in ordered_pairs[:L_contemp]]

    # Possibly cyclic lagged random pairs of links 
    # where we remove already chosen contemp links
    # Take up to (!) L_contemp links
    unordered_pairs = list(itertools.permutations(range(N), 2))
    unordered_pairs = list(set(unordered_pairs) - set(ordered_pairs[:L_contemp]))
    random_state.shuffle(unordered_pairs)
    lagged_links = [(causal_order[pair[0]], causal_order[pair[1]]) 
                    for pair in unordered_pairs[:L_lagged]]

    chosen_links = lagged_links + contemp_links

    # Populate links
    for (i, j) in chosen_links:

        # Choose lag
        if (i, j) in contemp_links:
            tau = 0
        else:
            tau = int(random_state.randint(1, max_lag+1))

        # Choose dependency
        c = float(random_state.choice(dependency_coeffs))
        if c != 0:
            func = random_state.choice(dependency_funcs)
            if func == 'linear':
                func = linear
            elif func == 'nonlinear':
                func = nonlinear

            links[j].append(((int(i), -tau), c, func))

    # Now generate noise functions
    # Either choose among pre-defined noise types or supply your own
    class NoiseModel:
        def __init__(self, mean=0., sigma=1.):
            self.mean = mean
            self.sigma = sigma
        def gaussian(self, T):
            # Get zero-mean unit variance gaussian distribution
            return self.mean + self.sigma*random_state_noise.randn(T)
        def weibull(self, T): 
            # Get zero-mean sigma variance weibull distribution
            a = 2
            mean = scipy.special.gamma(1./a + 1)
            variance = scipy.special.gamma(2./a + 1) - scipy.special.gamma(1./a + 1)**2
            return self.mean + self.sigma*(random_state_noise.weibull(a=a, size=T) - mean)/np.sqrt(variance)
        def uniform(self, T): 
            # Get zero-mean sigma variance uniform distribution
            mean = 0.5
            variance = 1./12.
            return self.mean + self.sigma*(random_state_noise.uniform(size=T) - mean)/np.sqrt(variance)

    noises = []
    for j in links:
        noise_dist = random_state.choice(noise_dists)
        noise_mean = random_state.choice(noise_means)
        noise_sigma = random_state.choice(noise_sigmas)

        if noise_dist in ['gaussian', 'weibull', 'uniform']:
            noise = getattr(NoiseModel(mean = noise_mean, sigma = noise_sigma), noise_dist)
        else:
            noise = noise_dist
        
        noises.append(noise)

    return links, noises

def check_stationarity(links):
    """Returns stationarity according to a unit root test.

    Assumes an at least asymptotically linear vector autoregressive process
    without contemporaneous links.

    Parameters
    ---------
    links : dict
        Dictionary of form {0:[((0, -1), coeff, func), ...], 1:[...], ...}.
        Also format {0:[(0, -1), ...], 1:[...], ...} is allowed.

    Returns
    -------
    stationary : bool
        True if VAR process is stationary.
    """

    N = len(links)
    # Check parameters
    max_lag = 0

    for j in range(N):
        for link_props in links[j]:
            var, lag = link_props[0]
            # coeff = link_props[1]
            # coupling = link_props[2]

            max_lag = max(max_lag, abs(lag))

    graph = np.zeros((N,N,max_lag))
    couplings = []

    for j in range(N):
        for link_props in links[j]:
            var, lag = link_props[0]
            coeff    = link_props[1]
            coupling = link_props[2]
            if abs(lag) > 0:
                graph[j,var,abs(lag)-1] = coeff
            couplings.append(coupling)

    stabmat = np.zeros((N*max_lag,N*max_lag))
    index = 0

    for i in range(0,N*max_lag,N):
        stabmat[:N,i:i+N] = graph[:,:,index]
        if index < max_lag-1:
            stabmat[i+N:i+2*N,i:i+N] = np.identity(N)
        index += 1

    eig = np.linalg.eig(stabmat)[0]

    if np.all(np.abs(eig) < 1.):
        stationary = True
    else:
        stationary = False

    return stationary
    # if len(eig) == 0:
    #     return stationary, 0.
    # else:
    #     return stationary, np.abs(eig).max()

class _Logger(object):
    """Class to append print output to a string which can be saved"""
    def __init__(self):
        self.terminal = sys.stdout
        self.log = ""       # open("log.dat", "a")

    def write(self, message):
        self.terminal.write(message)
        self.log += message  # .write(message)


if __name__ == '__main__':
    
    ## Generate some time series from a structural causal process
    def lin_f(x): return x
    def nonlin_f(x): return (x + 5. * x**2 * np.exp(-x**2 / 20.))

    links, noises = generate_structural_causal_process(seed=1, noise_seed=1)
    
    # data, nonstat = structural_causal_process(links, seed=1,
    #          T=10, noises=noises)

    data, nonstat = structural_causal_process_ensemble(realizations=2, ensemble_seed=0, 
        links=links, T=2, noises=noises)

    print(data)
    print(data.shape)


    # links = {0: [((0, -1), 0.9, lin_f)],
    #          1: [((1, -1), 0.8, lin_f), ((0, -1), 0.3, nonlin_f)],
    #          2: [((2, -1), 0.7, lin_f), ((1, 0), -0.2, lin_f)],
    #          }
    # noises = [np.random.randn, np.random.randn, np.random.randn]

    # data, nonstat = structural_causal_process(links,
    #  T=100, noises=noises)

